import os

from flask import Flask
from werkzeug.middleware.profiler import ProfilerMiddleware

import flask_camp


app = Flask(__name__, static_folder=None)
app.config.update({"SECRET_KEY": "not very secret", "SQLALCHEMY_TRACK_MODIFICATIONS": False})
app.wsgi_app = ProfilerMiddleware(
    app.wsgi_app,
    restrictions=[
        os.path.dirname(flask_camp.__file__),
    ],
)

api = flask_camp.RestApi(app=app)

with app.app_context():
    api.database.create_all()



from tests.unit_tests.utils import BaseTest


class Test_ETag(BaseTest):
    def test_main(self, user):
        self.login_user(user)

        v1 = self.create_document(data={"value": "42"}).json["document"]
        r = self.get_document(v1)
        assert "ETag" in r.headers, r.headers
        assert self.api.memory_cache.get_document(v1["id"])["timestamp"] == v1["timestamp"]
        etag = r.headers["ETag"]

        self.get_document(v1, headers={"If-None-Match": etag}, expected_status=304)
        self.get_document(v1, headers={"If-None-Match": "not-the-good-hash"}, expected_status=200)

        # a modification remove the document from the cache
        v2 = self.modify_document(v1, data="12").json["document"]
        assert self.api.memory_cache.get_document(v1["id"]) is None

        # but a get recompute it in the cache
        self.get_document(v1)
        assert self.api.memory_cache.get_document(v1["id"]) is not None
        assert self.api.memory_cache.get_document(v1["id"])["timestamp"] == v2["timestamp"]

        r = self.get_document(v1, headers={"If-None-Match": etag}, expected_status=200)

    def test_memcache_failure(self, user):
        self.login_user(user)

        doc = self.create_document().json["document"]
        self.api.memory_cache.delete_document(doc["id"])
        self.get_document(doc, expected_status=200)



# request tags
# get changes on document with some tag
# get changes on document with some tag value
# get changes on document with some tag for user X
# get changes on document with some tag value for user X

from tests.unit_tests.utils import BaseTest


def assert_tag(tag, user, document_id, name, value):
    assert len(tag) == 4
    assert tag["user_id"] == user.id, tag
    assert tag["document_id"] == document_id, tag
    assert tag["name"] == name, tag
    assert tag["value"] == value, tag


class Test_UserTag(BaseTest):
    def test_not_anonymous(self):
        self.post("/user_tags", expected_status=403)

    def test_add_modify_delete(self, user):
        self.login_user(user)

        doc = self.create_document().json["document"]

        self.add_user_tag("x", doc, expected_status=200)

        r = self.get_user_tags()
        assert_tag(r.json["user_tags"][0], user, doc["id"], "x", None)

        self.add_user_tag("y", doc, "6a", expected_status=200)

        r = self.get_user_tags()
        assert_tag(r.json["user_tags"][0], user, doc["id"], "x", None)
        assert_tag(r.json["user_tags"][1], user, doc["id"], "y", "6a")

        r = self.add_user_tag("x", doc, "6b", expected_status=200)
        r = self.get_user_tags()
        assert_tag(r.json["user_tags"][0], user, doc["id"], "x", "6b")
        assert_tag(r.json["user_tags"][1], user, doc["id"], "y", "6a")

        r = self.remove_user_tag("x", document=doc, expected_status=200)

        r = self.get_user_tags()
        assert_tag(r.json["user_tags"][0], user, doc["id"], "y", "6a")

    def test_get_tags(self, user, user_2):

        self.login_user(user)
        doc1 = self.create_document().json["document"]
        doc2 = self.create_document().json["document"]

        self.add_user_tag("t1", doc1)
        self.add_user_tag("t2", doc1)
        self.add_user_tag("t1", doc2)
        self.add_user_tag("t2", doc2)

        self.logout_user()
        self.login_user(user_2)

        self.add_user_tag("t1", doc1)
        self.add_user_tag("t2", doc1)
        self.add_user_tag("t1", doc2)
        self.add_user_tag("t2", doc2)

        r = self.get_user_tags()
        assert r.json["count"] == 8

        r = self.get_user_tags(user=user)
        assert r.json["count"] == 4

        r = self.get_user_tags(document=doc1)
        assert r.json["count"] == 4

        r = self.get_user_tags(name="t1")
        assert r.json["count"] == 4

        r = self.get_user_tags(user=user, document=doc1)
        assert r.json["count"] == 2

        r = self.get_user_tags(document=doc1, name="t1")
        assert r.json["count"] == 2

        r = self.get_user_tags(user=user, name="t1")
        assert r.json["count"] == 2

        r = self.get_user_tags(user=user, document=doc1, name="t1")
        assert r.json["count"] == 1

    def test_get_documents(self, user, user_2):
        self.login_user(user)
        doc1 = self.create_document().json["document"]
        doc2 = self.create_document().json["document"]
        doc3 = self.create_document().json["document"]

        self.add_user_tag("t1", doc1, "6a")
        self.add_user_tag("t1", doc2)
        self.add_user_tag("t2", doc3)

        self.logout_user()

        self.login_user(user_2)
        self.add_user_tag("t1", doc1)

        r = self.get_documents(tag_name="t1")
        assert r.json["count"] == 2

        r = self.get_documents(tag_name="t1", tag_user=user_2)
        assert r.json["count"] == 1

        r = self.get_documents(tag_name="t1", tag_value="6a")
        assert r.json["count"] == 1

    def test_errors(self, user):
        self.login_user(user)

        doc = self.create_document().json["document"]

        self.remove_user_tag(name="x", document=doc, expected_status=404)
        self.get_user_tags(limit=101, expected_status=400)
        self.get_documents(limit=101, expected_status=400)



from sqlalchemy import Column, Integer
from flask_camp._services._database import database


class BaseModel(database.Model):  # pylint: disable=too-few-public-methods
    __abstract__ = True  # tells SQLAlchemy that this model should not be created in the database

    id = Column(Integer, primary_key=True, index=True)

    @classmethod
    def get(cls, with_for_update=False, **kwargs):
        query = cls.query.filter_by(**kwargs)

        if with_for_update:
            query = query.with_for_update()

        return query.first()



import json
import time

from flask import request, Response
from flask_login import current_user
from werkzeug.exceptions import NotFound, Forbidden, Conflict, BadRequest

from flask_camp._schemas import schema
from flask_camp._utils import get_cooked_document, get_document, cook, current_api
from flask_camp.models._document import Document, DocumentVersion
from flask_camp._services._security import allow

rule = "/document/<int:document_id>"


class EditConflict(Conflict):
    def __init__(self, your_version, last_version):
        super().__init__("A new version exists")
        self.data = {
            "last_version": last_version,
            "your_version": your_version,
        }


@allow("anonymous", "authenticated", allow_blocked=True)
def get(document_id):
    """Get a document"""
    document_as_dict = get_cooked_document(document_id)  # it handles not found

    if document_as_dict.get("redirect_to"):
        return Response(
            headers={"Location": f"/document/{document_as_dict['redirect_to']}"},
            content_type="application/json",
            response=json.dumps({"status": "ok", "document": document_as_dict}),
            status=301,
        )

    response = Response(
        response=json.dumps({"status": "ok", "document": document_as_dict}),
        content_type="application/json",
    )

    response.add_etag()
    response.make_conditional(request)

    return response


@allow("authenticated")
@schema("modify_document.json")
def post(document_id):
    """add a new version to a document"""

    document = Document.get(id=document_id, with_for_update=True)

    if document is None:
        raise NotFound()

    if document.protected and not current_user.is_moderator:
        raise Forbidden("The document is protected")

    if document.is_redirection:
        raise BadRequest("The document is a redirection")

    body = request.get_json()

    comment = body["comment"]
    data = body["document"]["data"]

    current_api.validate_user_schemas(body["document"])

    version_id = body["document"]["version_id"]

    last_version = document.as_dict()

    if last_version["version_id"] != version_id:
        raise EditConflict(last_version=last_version, your_version=body["document"])

    version = DocumentVersion(
        document=document,
        user=current_user,
        comment=comment,
        data=data,
    )

    current_api.database.session.add(version)
    document.last_version = version

    document.associated_ids = current_api.get_associated_ids(version.as_dict())

    assert _RACE_CONDITION_TESTING()

    current_api.database.session.flush()

    current_api.before_document_save(document)

    current_api.database.session.commit()

    document.clear_memory_cache()

    return {"status": "ok", "document": cook(version.as_dict())}


@allow("authenticated")
@schema("action_with_comment.json")
def delete(document_id):
    """Delete a document"""
    document_as_dict = get_document(document_id)

    if not current_api.user_can_delete and not current_user.is_admin:
        raise Forbidden()

    current_api.before_document_delete(current_user, document_as_dict)

    document = Document.get(id=document_id)
    current_api.database.session.delete(document)

    current_api.add_log("delete_document", document=document, comment=request.get_json()["comment"])
    current_api.database.session.commit()

    document.clear_memory_cache()

    return {"status": "ok"}


def _RACE_CONDITION_TESTING():
    if "rc_sleep" in request.args:
        rc_sleep = float(request.args["rc_sleep"])
        time.sleep(rc_sleep)

    return True



"""validate the user email with the validation token"""

from flask import request
from sqlalchemy.exc import IntegrityError
from werkzeug.exceptions import BadRequest, NotFound

from flask_camp._schemas import schema
from flask_camp._utils import current_api
from flask_camp._services._security import allow
from flask_camp.models._user import User as UserModel

rule = "/validate_email"


@allow("admin")
def get():
    """Resend validation mail to a user. Only admin can do this request"""
    name = request.args.get("name", "")

    if not name:
        raise BadRequest()

    user = UserModel.get(name=name)
    if not user:
        raise NotFound()

    user.send_account_creation_mail()

    return {"status": "ok"}


@allow("anonymous", "authenticated")
@schema("validate_email.json")
def post():
    """Validate an email"""
    data = request.get_json()
    user = UserModel.get(name=data["name"])

    if user is None:
        raise NotFound()

    user.validate_email(data["token"])

    try:
        current_api.database.session.commit()
    except IntegrityError as e:
        raise BadRequest("A user still exists with this email") from e

    return {"status": "ok"}



from tests.unit_tests.utils import BaseTest


class Test_CurrentUser(BaseTest):
    def test_anonymous(self):
        self.get_current_user(expected_status=403)

    def test_basic(self, user):
        self.login_user(user)
        r = self.get_current_user(expected_status=200)

        assert "user" in r.json



import pytest

from flask_camp.models import User
from flask_camp._schemas import schema
from tests.unit_tests.utils import BaseTest


def test_missing_schema():
    with pytest.raises(FileNotFoundError):
        schema("Idonotexists")


class Test_Vuln(BaseTest):
    def test_vuln(self, user, user_2):
        @self.app.route("/__testing/vuln/<int:user_id>", methods=["GET"])
        def testing_vuln(user_id):
            """Calling this method without being authentified as user_id mys raise a Forbidden response"""
            return User.get(id=user_id).as_dict(include_personal_data=True)

        self.get(f"/__testing/vuln/{user.id}", expected_status=403)

        self.login_user(user)
        self.get(f"/__testing/vuln/{user_2.id}", expected_status=403)

        self.get(f"/__testing/vuln/{user.id}", expected_status=200)


class Test_Errors(BaseTest):
    def test_no_body(self, user):
        self.login_user(user)

        r = self.put("/documents", data="null", expected_status=400)
        assert r.json is not None

        r = self.put("/documents", data="null", headers={"Content-Type": "application/json"}, expected_status=400)
        assert r.json is not None
        assert r.json["description"] == "None is not of type 'object' on instance ", r.json

    def test_main(self):
        r = self.get("/do_not_exists", expected_status=404)
        assert r.json is not None
        assert r.json["status"] == "error"

        r = self.delete("/healthcheck", expected_status=405)
        assert r.json is not None
        assert r.json["status"] == "error"



from tests.unit_tests.utils import BaseTest


class Test_GetVersion(BaseTest):
    def test_errors(self):
        self.get_version(42, expected_status=404)

    def test_main(self, user, moderator):
        self.login_user(user)

        v0 = self.create_document().json["document"]
        v1 = self.modify_document(v0, data={"value": "43"}).json["document"]

        self.logout_user()

        r = self.get_version(v0, expected_status=200)
        assert r.json["document"]["data"] == {}

        r = self.get_version(v1, expected_status=200)
        assert r.json["document"]["data"] == {"value": "43"}

        self.login_user(moderator)

        self.hide_version(v0)

        r = self.get_version(v0, expected_status=200)
        assert r.json["document"]["data"] == {}

        self.login_user(user)

        r = self.get_version(v0, data_should_be_present=False, expected_status=200)
        assert r.json["document"]["hidden"] is True
        assert "data" not in r.json["document"]


class Test_DeleteVersion(BaseTest):
    def test_main(self, admin):
        self.login_user(admin)

        v0 = self.create_document().json["document"]

        v1 = self.modify_document(v0, data={"value": "43"}).json["document"]
        self.modify_document(v1, data={"value": "43"})

        r = self.get_versions(document=v0)
        assert r.json["count"] == 3

        self.logout_user()
        self.login_user(admin)

        self.delete_version(v1, expected_status=200)

        r = self.get_versions(document=v0)
        assert r.json["count"] == 2

    def test_not_the_last_one(self, admin):
        self.login_user(admin)

        v0 = self.create_document().json["document"]
        r = self.delete_version(v0, expected_status=400)
        assert r.json["description"] == "Can't delete last version of a document"

    def test_rights(self, user):
        self.login_user(user)

        v0 = self.create_document().json["document"]
        self.delete_version(v0, expected_status=403)

    def test_not_found(self, admin):
        self.login_user(admin)

        self.delete_version(42, expected_status=404)

    def test_bad_format(self, admin):
        self.login_user(admin)

        self.delete("/version/200", json={"commentt": "toto"}, expected_status=400)
        self.delete("/version/200", expected_status=400)



""" Views related to account operations """

from flask import request

from flask_camp._schemas import schema
from flask_camp._utils import current_api
from flask_camp._services._security import allow
from flask_camp.models._user import User as UserModel


rule = "/reset_password"


@allow("anonymous", "authenticated")
@schema("reset_password.json")
def post():
    """Send an email with a login token to this user"""
    email = request.get_json()["email"]

    user = UserModel.get(_email=email)

    if user is None:  # do not let hacker crawl our base
        fake_user = UserModel()
        fake_user.set_login_token()

        return {"status": "ok", "expiration_date": fake_user.login_token_expiration_date.isoformat()}

    user.set_login_token()

    current_api.database.session.commit()

    user.send_login_token_mail()

    return {"status": "ok", "expiration_date": user.login_token_expiration_date.isoformat()}



from flask import request
from flask_login import current_user
from werkzeug.exceptions import BadRequest, NotFound

from flask_camp._schemas import schema
from flask_camp._utils import current_api
from flask_camp._services._security import allow
from flask_camp.models._user_tag import UserTag

rule = "/user_tags"


def _build_filters(**kwargs):
    return {key: value for key, value in kwargs.items() if value is not None}


@allow("anonymous", "authenticated")
def get():
    """Get user tag list"""
    limit = request.args.get("limit", default=100, type=int)
    offset = request.args.get("offset", default=0, type=int)

    if not 0 <= limit <= 100:
        raise BadRequest("Limit can't be lower than 0 or higher than 100")

    filters = _build_filters(
        user_id=request.args.get("user_id", default=None, type=int),
        document_id=request.args.get("document_id", default=None, type=int),
        name=request.args.get("name", default=None, type=str),
        value=request.args.get("value", default=None, type=str),
    )

    query = UserTag.query

    if len(filters) != 0:
        query = query.filter_by(**filters)

    count = query.count()
    query = query.order_by(UserTag.id).offset(offset).limit(limit)

    return {
        "status": "ok",
        "count": count,
        "user_tags": [tag.as_dict() for tag in query],
    }


@allow("authenticated", allow_blocked=True)
@schema("modify_user_tag.json")
def post():
    """create/modify an user tag"""
    data = request.get_json()

    document_id = data["document_id"]
    name = data["name"]
    value = data.get("value", None)

    tag = UserTag.get(name=name, document_id=document_id, user_id=current_user.id)
    if tag is None:
        tag = UserTag(name=name, document_id=document_id, user_id=current_user.id)
        current_api.database.session.add(tag)

    tag.value = value

    current_api.database.session.commit()

    return {"status": "ok", "user_tag": tag.as_dict()}


@allow("authenticated", allow_blocked=True)
@schema("delete_user_tag.json")
def delete():
    """Delete an user tag"""
    data = request.get_json()

    document_id = data["document_id"]
    name = data["name"]

    tag = UserTag.get(name=name, document_id=document_id, user_id=current_user.id)

    if not tag:
        raise NotFound()

    current_api.database.session.delete(tag)
    current_api.database.session.commit()

    return {"status": "ok"}



from copy import deepcopy


class ClientInterface:
    @staticmethod
    def _get_user_id(user):
        return user if isinstance(user, int) else user["id"] if isinstance(user, dict) else user.id

    @staticmethod
    def _get_document_id(document):
        return document if isinstance(document, int) else document["id"] if isinstance(document, dict) else document.id

    def get(self, url, **kwargs):
        raise NotImplementedError()

    def post(self, url, **kwargs):
        raise NotImplementedError()

    def put(self, url, **kwargs):
        raise NotImplementedError()

    def delete(self, url, **kwargs):
        raise NotImplementedError()

    ###########################################################################################

    def create_user(self, name, email, password, **kwargs):
        json = {"name": name, "email": email, "password": password} | kwargs.pop("json", {})
        return self.put("/users", json=json, **kwargs)

    def validate_email(self, user, token, **kwargs):
        name = user if isinstance(user, str) else user["name"] if isinstance(user, dict) else user.name
        return self.post("/validate_email", json={"name": name, "token": token}, **kwargs)

    def resend_email_validation(self, user, **kwargs):
        name = user if isinstance(user, str) else user["name"] if isinstance(user, dict) else user.name
        return self.get("/validate_email", params={"name": name}, **kwargs)

    def reset_password(self, email, **kwargs):
        return self.post("/reset_password", json={"email": email}, **kwargs)

    def login_user(self, user, password=None, token=None, **kwargs):
        name = user if isinstance(user, str) else user["name"] if isinstance(user, dict) else user.name

        payload = {"name": name}
        if token is not None:
            payload["token"] = token
        else:
            payload["password"] = password

        return self.post("/login", json=payload | kwargs.pop("json", {}), **kwargs)

    def get_current_user(self, **kwargs):
        return self.get("/current_user", **kwargs)

    def logout_user(self, **kwargs):
        return self.delete("/login", **kwargs)

    def get_user(self, user, **kwargs):
        user_id = self._get_user_id(user)
        return self.get(f"/user/{user_id}", **kwargs)

    def get_users(self, limit=None, **kwargs):
        params = {}

        if limit is not None:
            params["limit"] = limit

        return self.get("/users", params=params | kwargs.pop("params", {}), **kwargs)

    def modify_user(
        self,
        user,
        password=None,
        token=None,
        new_password=None,
        email=None,
        ui_preferences=None,
        **kwargs,
    ):
        user_id = self._get_user_id(user)
        json = {}

        if password is not None:
            json["password"] = password

        if token is not None:
            json["token"] = token

        if new_password is not None:
            json["new_password"] = new_password

        if email is not None:
            json["email"] = email

        if ui_preferences is not None:
            json["ui_preferences"] = ui_preferences

        return self.post(f"/user/{user_id}", json=json | kwargs.pop("json", {}), **kwargs)

    def rename_user(self, user, name, comment, **kwargs):
        user_id = self._get_user_id(user)

        return self.post(
            f"/rename_user/{user_id}",
            json={"name": name, "comment": comment} | kwargs.pop("json", {}),
            **kwargs,
        )

    def add_user_role(self, user, role, comment, **kwargs):
        user_id = self._get_user_id(user)

        return self.post(
            f"/roles/{user_id}",
            json={"role": role, "comment": comment} | kwargs.pop("json", {}),
            **kwargs,
        )

    def remove_user_role(self, user, role, comment, **kwargs):
        user_id = self._get_user_id(user)

        return self.delete(
            f"/roles/{user_id}",
            json={"role": role, "comment": comment} | kwargs.pop("json", {}),
            **kwargs,
        )

    def get_user_roles(self, user, **kwargs):
        user_id = self._get_user_id(user)
        return self.get(f"/roles/{user_id}", **kwargs)

    def create_document(self, data, comment, **kwargs):
        return self.put(
            "/documents",
            json={"comment": comment, "document": {"data": data}} | kwargs.pop("json", {}),
            **kwargs,
        )

    def get_documents(
        self,
        limit=None,
        offset=None,
        tag_name=None,
        tag_user=None,
        tag_value=None,
        **kwargs,
    ):
        params = {}

        if tag_name is not None:
            params["tag_name"] = tag_name

        if tag_user is not None:
            params["tag_user_id"] = self._get_user_id(tag_user)

        if tag_value is not None:
            params["tag_value"] = tag_value

        if limit is not None:
            params["limit"] = limit

        if offset is not None:
            params["offset"] = offset

        return self.get("/documents", params=params | kwargs.pop("params", {}), **kwargs)

    def get_document(self, document, **kwargs):
        document_id = document if isinstance(document, int) else document["id"]
        return self.get(f"/document/{document_id}", **kwargs)

    def get_version(self, version, **kwargs):
        version_id = version if isinstance(version, int) else version["version_id"]
        return self.get(f"/version/{version_id}", **kwargs)

    def get_versions(self, document=None, limit=None, user=None, tag_name=None, tag_user=None, **kwargs):
        params = {}

        if document is not None:
            params["document_id"] = document["id"]

        if user is not None:
            params["user_id"] = self._get_user_id(user)

        if tag_name is not None:
            params["tag_name"] = tag_name

        if tag_user is not None:
            params["tag_user_id"] = self._get_user_id(tag_user)

        if limit is not None:
            params["limit"] = limit

        return self.get("/versions", params=params | kwargs.pop("params", {}), **kwargs)

    def modify_document(self, document, comment, data, **kwargs):
        document_id = self._get_document_id(document)
        new_version = deepcopy(document)
        new_version["data"] = data

        return self.post(
            f"/document/{document_id}",
            json={"comment": comment, "document": new_version} | kwargs.pop("json", {}),
            **kwargs,
        )

    def hide_version(self, version, comment, **kwargs):
        version_id = version if isinstance(version, int) else version["version_id"]
        return self.post(
            f"/version/{version_id}",
            json={"comment": comment, "hidden": True} | kwargs.pop("json", {}),
            **kwargs,
        )

    def unhide_version(self, version, comment, **kwargs):
        version_id = version if isinstance(version, int) else version["version_id"]
        return self.post(
            f"/version/{version_id}",
            json={"comment": comment, "hidden": False} | kwargs.pop("json", {}),
            **kwargs,
        )

    def protect_document(self, document, comment, **kwargs):
        document_id = document if isinstance(document, int) else document["id"]
        return self.post(
            f"/protect_document/{document_id}",
            json={"comment": comment, "protected": True},
            **kwargs,
        )

    def unprotect_document(self, document, comment, **kwargs):
        document_id = document if isinstance(document, int) else document["id"]

        return self.post(
            f"/protect_document/{document_id}",
            json={"comment": comment, "protected": False} | kwargs.pop("json", {}),
            **kwargs,
        )

    def block_user(self, user, comment, **kwargs):
        user_id = self._get_user_id(user)
        return self.post(
            f"/block_user/{user_id}",
            json={"comment": comment, "blocked": True} | kwargs.pop("json", {}),
            **kwargs,
        )

    def unblock_user(self, user, comment, **kwargs):
        user_id = self._get_user_id(user)
        return self.post(
            f"/block_user/{user_id}",
            json={"comment": comment, "blocked": False} | kwargs.pop("json", {}),
            **kwargs,
        )

    def delete_version(self, version, comment, **kwargs):
        version_id = version if isinstance(version, int) else version["version_id"]
        return self.delete(f"/version/{version_id}", json={"comment": comment} | kwargs.pop("json", {}), **kwargs)

    def delete_document(self, document, comment, **kwargs):
        document_id = document if isinstance(document, int) else document["id"]
        return self.delete(f"/document/{document_id}", json={"comment": comment} | kwargs.pop("json", {}), **kwargs)

    def get_user_tags(self, limit=None, offset=None, user=None, document=None, name=None, **kwargs):
        params = {}

        if document is not None:
            params["document_id"] = self._get_document_id(document)

        if user is not None:
            params["user_id"] = self._get_user_id(user)

        if name is not None:
            params["name"] = name

        if limit is not None:
            params["limit"] = limit

        if limit is not None:
            params["offset"] = offset

        return self.get("/user_tags", params=params | kwargs.pop("params", {}), **kwargs)

    def add_user_tag(self, name, document, value=None, **kwargs):
        json = {"name": name, "document_id": self._get_document_id(document)}

        if value is not None:
            json["value"] = value

        return self.post("/user_tags", json=json | kwargs.pop("json", {}), **kwargs)

    def remove_user_tag(self, name, document, **kwargs):
        return self.delete(
            "/user_tags",
            json={
                "name": name,
                "document_id": self._get_document_id(document),
            }
            | kwargs.pop("json", {}),
            **kwargs,
        )

    def merge_documents(self, document_to_merge, document_destination, comment, **kwargs):
        json = {
            "document_to_merge": document_to_merge["id"],
            "document_destination": document_destination["id"],
            "comment": comment,
        } | kwargs.pop("params", {})

        return self.post("/merge", json=json, **kwargs)

    def get_logs(self, limit=None, **kwargs):
        params = {}

        if limit is not None:
            params["limit"] = limit

        return self.get("/logs", params=params | kwargs.pop("params", {}), **kwargs)



from flask import request
from sqlalchemy import select, func
from werkzeug.exceptions import BadRequest

from flask_camp._schemas import schema
from flask_camp._utils import get_cooked_document, cook, current_api
from flask_camp._services._security import allow
from flask_camp.models._document import Document

rule = "/documents"


@allow("anonymous", "authenticated", allow_blocked=True)
def get():
    """Get a list of documents"""

    limit = request.args.get("limit", default=30, type=int)
    offset = request.args.get("offset", default=0, type=int)

    if not 0 <= limit <= 100:
        raise BadRequest("Limit can't be lower than 0 or higher than 100")

    def make_query(base_query):

        query = base_query

        query = query.where(Document.redirect_to.is_(None))

        tag_filters_args = {
            "user_id": request.args.get("tag_user_id", default=None, type=int),
            "name": request.args.get("tag_name", default=None, type=str),
            "value": request.args.get("tag_value", default=None, type=str),
        }

        tag_filters_args = {k: v for k, v in tag_filters_args.items() if v is not None}

        if len(tag_filters_args) != 0:
            query = query.where(Document.user_tags.any(**tag_filters_args))

        query = current_api.update_search_query(query)

        return current_api.database.session.execute(query)

    count = make_query(select(func.count(Document.id)))
    document_ids = make_query(select(Document.id).limit(limit).offset(offset).order_by(Document.id.asc()))

    documents = [get_cooked_document(row[0]) for row in document_ids]
    return {"status": "ok", "documents": documents, "count": list(count)[0][0]}


@allow("authenticated")
@schema("create_document.json")
def put():
    """create an document"""
    body = request.get_json()

    current_api.validate_user_schemas(body["document"])

    document = Document.create(
        comment=body["comment"],
        data=body["document"]["data"],
    )

    current_api.database.session.flush()

    current_api.before_document_save(document)

    current_api.database.session.commit()

    document.clear_memory_cache()

    return {"status": "ok", "document": cook(document.last_version.as_dict())}



from tests.unit_tests.utils import BaseTest


class Test_Protection(BaseTest):
    def test_errors(self, moderator):
        self.login_user(moderator)

        document = self.create_document().json["document"]

        self.protect_document(42, expected_status=404)
        self.unprotect_document(42, expected_status=404)

        self.protect_document(document)
        self.protect_document(document, expected_status=400)

        self.unprotect_document(document)
        self.unprotect_document(document, expected_status=400)

    def test_typical_scenario(self, user, moderator):
        self.login_user(user)
        v0 = self.create_document().json["document"]
        document = v0  # more clear

        # try to protect a doc without being an moderator
        r = self.protect_document(document, expected_status=403)
        self.logout_user()

        # protect doc
        self.login_user(moderator)
        r = self.protect_document(document)

        r = self.get_document(document)
        assert r.json["document"]["protected"] is True
        self.logout_user()

        self.login_user(user)
        self.protect_document(document, expected_status=403)  # unprotect doc without being an moderator
        self.modify_document(v0, expected_status=403)  # edit protected doc without being an moderator
        self.logout_user()

        self.login_user(moderator)
        v1 = self.modify_document(v0, data={"value": "43"}, expected_status=200).json["document"]  # edit protected doc
        self.unprotect_document(document, expected_status=200)  # unprotect doc

        r = self.get_document(document)
        assert r.json["document"]["protected"] is False
        self.logout_user()

        # edit deprotected doc
        self.login_user(user)
        v2 = self.modify_document(v1, data={"value": "44"}, expected_status=200).json["document"]
        assert r.json["document"]["protected"] is False

        # try to hack
        v2["protected"] = True
        v3 = self.modify_document(v2, data={"value": "45"}, expected_status=200).json["document"]
        assert v3["protected"] is False



from flask import request
from werkzeug.exceptions import NotFound

from flask_camp._schemas import schema
from flask_camp._utils import current_api
from flask_camp.models._user import User
from flask_camp._services._security import allow

rule = "/rename_user/<int:user_id>"


@allow("moderator")
@schema("rename_user.json")
def post(user_id):
    """rename an user"""

    user = User.get(id=user_id, with_for_update=True)

    if user is None:
        raise NotFound()

    data = request.get_json()
    user.name = data["name"]

    current_api.add_log(action="Rename user", comment=data["comment"], target_user=user)
    current_api.database.session.commit()

    return {"status": "ok"}



"""user tag is ..."""

from sqlalchemy import Column, ForeignKey, Integer, String, UniqueConstraint
from sqlalchemy.orm import relationship

from flask_camp.models._base_model import BaseModel
from flask_camp.models._user import User


class UserTag(BaseModel):
    __tablename__ = "user_tag"

    user_id = Column(Integer, ForeignKey(User.id), index=True, nullable=False)
    user = relationship(User, foreign_keys=[user_id])

    document_id = Column(Integer, ForeignKey("document.id"), index=True, nullable=False)
    document = relationship("Document", foreign_keys=[document_id])

    name = Column(String(16), index=True, nullable=False)
    value = Column(String(32), index=True)

    __table_args__ = (UniqueConstraint("user_id", "document_id", "name", name="_user_tag_uc"),)

    def as_dict(self):
        return {
            "user_id": self.user_id,
            "document_id": self.document_id,
            "name": self.name,
            "value": self.value,
        }



import collections

from flask import current_app

from flask_camp._utils import current_api
from flask_camp._services._security import allow

rule = "/"


@allow("anonymous", "authenticated", allow_blocked=True)
def get():
    """Display the possible route for this API"""

    site_map = collections.defaultdict(dict)

    for url_rule in current_app.url_map.iter_rules():
        for method in url_rule.methods:
            if method not in ["OPTIONS", "HEAD"]:
                view_func = current_app.view_functions.get(url_rule.endpoint, None)
                doc = "" if view_func is None else view_func.__doc__
                site_map[url_rule.rule][method] = {"description": doc}

                if url_rule.rule in current_api._rate_limits and method in current_api._rate_limits[url_rule.rule]:
                    site_map[url_rule.rule][method]["rate_limit"] = current_api._rate_limits[url_rule.rule][method]

                if hasattr(view_func, "allowed_roles"):
                    site_map[url_rule.rule][method]["allowed_roles"] = list(view_func.allowed_roles)

                if hasattr(view_func, "allow_blocked"):
                    site_map[url_rule.rule][method]["allow_blocked"] = view_func.allow_blocked

    return site_map



from tests.unit_tests.utils import BaseTest


class Test_DocumentVersions(BaseTest):
    def test_main(self):
        r = self.get_versions(expected_status=200)
        assert r.json["versions"] == []

    def test_simple(self, user):
        self.login_user(user)

        doc1 = self.create_document(data={"value": "doc_1/v1"}).json["document"]
        doc2 = self.create_document(data={"value": "doc_2/v1"}).json["document"]

        self.modify_document(doc1, data={"value": "doc_1/v2"})
        self.modify_document(doc2, data={"value": "doc_2/v2"})

        r = self.get_versions(document=doc1, expected_status=200)
        history = r.json
        assert history["count"] == len(history["versions"]) == 2
        assert history["versions"][0]["version_id"] > history["versions"][1]["version_id"]
        assert history["versions"][0]["timestamp"] > history["versions"][1]["timestamp"]
        assert history["versions"][0]["id"] == history["versions"][1]["id"] == doc1["id"]
        assert history["versions"][0]["data"] == {"value": "doc_1/v2"}
        assert history["versions"][1]["data"] == {"value": "doc_1/v1"}

    def test_user_filter(self, user, user_2):

        self.login_user(user)
        doc = self.create_document(data={"value": "x"}).json["document"]
        self.modify_document(doc, data={"value": "y"})
        self.logout_user()

        self.login_user(user_2)
        doc = self.create_document(data={"value": "x"}).json["document"]
        self.modify_document(doc, data={"value": "y"})
        self.logout_user()

        r = self.get_versions(user=user, expected_status=200)

        history = r.json
        assert history["count"] == len(history["versions"]) == 2

    def test_user_tag_filter(self, user, user_2):
        self.login_user(user)
        doc1 = self.create_document().json["document"]
        doc2 = self.create_document().json["document"]
        self.add_user_tag("follow_list", doc1)
        self.logout_user()

        self.login_user(user_2)
        doc3 = self.create_document().json["document"]
        self.add_user_tag("follow_list", doc2)
        self.add_user_tag("other", doc3)
        self.logout_user()

        r = self.get_versions()
        assert r.json["count"] == 3

        r = self.get_versions(tag_name="follow_list")
        assert r.json["count"] == 2

        r = self.get_versions(tag_name="follow_list", tag_user=user)
        assert r.json["count"] == 1
        assert r.json["versions"][0]["id"] == doc1["id"]

        r = self.get_versions(tag_name="follow_list", tag_user=user_2)
        assert r.json["count"] == 1
        assert r.json["versions"][0]["id"] == doc2["id"]



import re
from tests.unit_tests.utils import BaseTest


class Test_UserCreation(BaseTest):
    def test_typical_scenario(self):
        name, email, password = "my_user", "a@b.c", "week password"

        with self.api.mail.record_messages() as outbox:
            r = self.create_user(name, email, password, expected_status=200)
            assert len(outbox) == 1
            assert outbox[0].subject == "Welcome to example.com"
            body = outbox[0].body
            token = re.sub(r"^(.*email_token=)", "", body)

        assert r.json["status"] == "ok"

        user = r.json["user"]

        assert len(user) == 5, user
        assert "id" in user
        assert "ui_preferences" in user
        assert user["blocked"] is False
        assert user["name"] == name
        assert user["roles"] == []

        self.validate_email(user=user, token=token, expected_status=200)

        # user should not be logged
        r = self.get_user(user)
        assert "email" not in r.json["user"]  # email is a private value

        r = self.get_user_roles(user)
        assert r.json["roles"] == []

        r = self.login_user(name, password, expected_status=200)

        assert len(r.json["user"]) == 6, r.json["user"]
        assert r.json["user"]["id"] == user["id"]
        assert r.json["user"]["blocked"] is False
        assert r.json["user"]["ui_preferences"] == {}
        assert r.json["user"]["name"] == name
        assert r.json["user"]["email"] == email
        assert r.json["user"]["roles"] == []

        self.logout_user()

    def test_errors_on_token_validation(self, unvalidated_user):

        r = self.login_user(unvalidated_user, expected_status=401)
        assert r.json["description"] == "User's email is not validated"

        r = self.post("/validate_email", json={"name": unvalidated_user.name}, expected_status=400)
        assert r.json["description"] == "'token' is a required property on instance "

        self.post(
            "/validate_email",
            json={"name": "not_the_name", "token": unvalidated_user._email_token},
            expected_status=404,
        )

        r = self.validate_email(unvalidated_user, token="not the good one", expected_status=401)
        assert r.json["description"] == "Token doesn't match"

        r = self.login_user(unvalidated_user, expected_status=401)
        assert r.json["description"] == "User's email is not validated"

        self.validate_email(unvalidated_user, token=unvalidated_user._email_token, expected_status=200)

        r = self.validate_email(unvalidated_user, token=unvalidated_user._email_token, expected_status=400)
        assert r.json["description"] == "There is no email to validate"

        r = self.login_user(unvalidated_user, expected_status=200)

    def test_login_errors(self, user):
        r = self.login_user("not_the_name", expected_status=401)
        assert r.json["description"] == "User [not_the_name] does not exists, or password is wrong"

        r = self.login_user(user, "not the password", expected_status=401)
        assert r.json["description"] == f"User [{user.name}] does not exists, or password is wrong"

        self.post("/login", json={"name": user.name}, expected_status=400)

    def test_logout_errors(self):
        self.delete("/login", expected_status=403)

    def test_notfound_errors(self, user):
        self.login_user(user)
        self.get("/user/42", expected_status=404)

    def test_anonymous_get(self, user):
        self.get_user(user, expected_status=200)

    def test_name_errors(self):
        email, password = "valid@email.com", "password"

        self.create_user("", email, password, expected_status=400)
        self.create_user("tailing_space ", email, password, expected_status=400)
        self.create_user(" starting_space", email, password, expected_status=400)
        self.create_user("abc", email, password, expected_status=400)  # too short
        self.create_user("@xxxx", email, password, expected_status=400)  # can't contains an @
        self.create_user("aaa@aaa", email, password, expected_status=400)  # same
        self.create_user("x" * 1000, email, password, expected_status=400)  # too long

    def test_email_error(self):
        name, password = "valid_name", "password"

        self.put("/users", json={"name": name, "email": None, "password": password}, expected_status=400)
        self.create_user(name, email="", password=password, expected_status=400)
        self.create_user(name, email="a.fr", password=password, expected_status=400)

    def test_create_while_logged(self, user):
        self.login_user(user)
        self.create_user(name="other", email="a@b.c", password="p", expected_status=400)

    def test_admin_can_resend_email(self, admin):
        user = self.create_user().json["user"]

        self.login_user(admin)

        with self.api.mail.record_messages() as outbox:
            self.resend_email_validation(user)
            assert len(outbox) == 1
            assert outbox[0].subject == "Welcome to example.com"
            body = outbox[0].body
            token = re.sub(r"^(.*email_token=)", "", body)

        self.logout_user()
        self.validate_email(user, token)

        self.login_user(user)

    def test_resend_validation_mail_errors(self, admin, moderator, user):
        new_user = self.create_user().json["user"]

        self.resend_email_validation(new_user, expected_status=403)

        self.login_user(user)
        self.resend_email_validation(new_user, expected_status=403)

        self.login_user(moderator)
        self.resend_email_validation(new_user, expected_status=403)

        self.login_user(admin)
        self.resend_email_validation(new_user, expected_status=200)

        self.get("/validate_email", params={"name": "not_the_name"}, expected_status=404)
        self.get("/validate_email", expected_status=400)


class Test_Errors(BaseTest):
    def test_mail_error(self):
        def raise_exception(*args, **kwargs):
            raise Exception("That was not expcted!")

        original_send = self.api.mail.send
        self.api.mail.send = raise_exception

        with self.api.mail.record_messages() as outbox:
            try:
                self.create_user(expected_status=200)
                assert len(outbox) == 0
            finally:
                self.api.mail.send = original_send


class Test_UserModification(BaseTest):
    rest_api_kwargs = {"user_roles": "bot,contributor"}

    def test_change_password(self, user):
        self.login_user(user)

        self.modify_user(user, new_password="p2", password="password")

        self.logout_user()
        self.login_user(user, "p1", expected_status=401)
        self.login_user(user, "p2", expected_status=200)

    def test_change_email(self, user):
        self.login_user(user)

        with self.api.mail.record_messages() as outbox:
            r = self.modify_user(user, email="other@email.com", password="password")
            assert len(outbox) == 1
            token = re.sub(r"^(.*email_token=)", "", outbox[0].body)

        self.logout_user()

        r = self.login_user(user)
        assert r.json["user"]["email"] == user._email  # not yet validated

        self.validate_email(user, token, expected_status=200)

        r = self.get_user(user, expected_status=200)
        assert r.json["user"]["email"] == "other@email.com", r.json

    def test_allowed_changes(self, user):
        self.login_user(user)

        new_values = {
            "name": "other_name",
            "ui_preferences": "UI",
        }

        self.post(f"/user/{user.id}", json=new_values)

        r = self.get_current_user()

        assert r.json["user"]["name"] == user.name
        assert r.json["user"]["roles"] == user.roles
        assert r.json["user"]["ui_preferences"] == new_values["ui_preferences"]

    def test_errors(self, admin, user, user_2):
        self.modify_user(user, ui_preferences="", expected_status=403)

        self.login_user(user)

        self.modify_user(user, new_password="13", expected_status=403)  # you must give the password
        self.modify_user(user, email="a@b.fr", expected_status=403)  # you must give the password

        self.modify_user(user, password="not the good pass", new_password="13", expected_status=403)
        self.modify_user(user, password="not the good pass", email="a@b.fr", expected_status=403)

        r = self.modify_user(user_2, new_password="p2", password="password", expected_status=403)
        assert r.json["description"] == "You can't modify this user"

        self.post(f"/user/{user.id}", json={"id": 12}, expected_status=400)

        self.login_user(admin)
        self.get_user_roles(42, expected_status=404)
        self.add_user_role(42, "admin", "test", expected_status=404)
        self.remove_user_role(42, "admin", "test", expected_status=404)

        self.add_user_role(user, "bot", "test", expected_status=200)
        self.add_user_role(user, "bot", "test", expected_status=400)

        self.remove_user_role(user, "bot", "test", expected_status=200)
        self.remove_user_role(user, "bot", "test", expected_status=400)

    def test_email_error(self, user):
        self.login_user(user)

        self.post(f"/user/{user.id}", json={"email": None, "password": "p"}, expected_status=400)
        self.modify_user(user, email="", password="password", expected_status=400)
        self.modify_user(user, email="a.fr", password="password", expected_status=400)


class Test_UserUniqueness(BaseTest):
    def test_username(self, user):
        r = self.create_user(user.name, "other@email.c", "password", expected_status=400)
        assert r.json["description"] == "A user still exists with this name"

    def test_email_at_creation(self, user):
        r = self.create_user("other_name", user._email, "password", expected_status=400)
        assert r.json["description"] == "A user still exists with this email"

    def test_email_at_modification(self, user, user_2):
        self.login_user(user)
        r = self.modify_user(user, password="password", email=user_2._email, expected_status=400)
        assert r.json["description"] == "A user still exists with this email"
        self.modify_user(user, password="password", email="mail@competition.fr", expected_status=200)
        self.logout_user()

        self.login_user(user_2)
        self.modify_user(user_2, password="password", email="mail@competition.fr", expected_status=200)
        self.logout_user()

    def test_do_not_validate_same_email(self):
        with self.api.mail.record_messages() as outbox:
            user_1 = self.create_user("user1", "a@b.c", "p").json["user"]
            token_1 = re.sub(r"^(.*email_token=)", "", outbox[0].body)

        with self.api.mail.record_messages() as outbox:
            user_2 = self.create_user("user2", "a@b.c", "p").json["user"]
            token_2 = re.sub(r"^(.*email_token=)", "", outbox[0].body)

        self.validate_email(user=user_1, token=token_1)
        r = self.validate_email(user=user_2, token=token_2, expected_status=400)
        assert r.json["description"] == "A user still exists with this email"


class Test_Logout(BaseTest):
    def test_main(self, user):
        self.logout_user(expected_status=403)
        self.login_user(user)
        self.logout_user()
        self.logout_user(expected_status=403)


class Test_Users(BaseTest):
    def test_main(self, admin, moderator, user):
        self.login_user(moderator)
        r = self.get_users().json

        assert r["count"] == 3
        assert r["users"][0]["id"] == user.id
        assert r["users"][0]["name"] == user.name
        assert r["users"][1]["id"] == moderator.id
        assert r["users"][1]["name"] == moderator.name
        assert r["users"][2]["id"] == admin.id
        assert r["users"][2]["name"] == admin.name

    def test_errors(self, moderator):
        self.get_users(expected_status=403)

        self.login_user(moderator)
        self.get_users(limit=1000, expected_status=400)



import json

from flask import request
from flask_login import current_user
from werkzeug.exceptions import Forbidden, NotFound

from flask_camp._schemas import schema
from flask_camp._utils import current_api
from flask_camp._services._security import allow
from flask_camp.models._user import User as UserModel


rule = "/user/<int:user_id>"


@allow("anonymous", "authenticated", allow_blocked=True)
def get(user_id):
    """Get an user"""
    user = UserModel.get(id=user_id)

    if user is None:
        raise NotFound()

    include_personal_data = False

    if current_user.is_authenticated:
        if user.id == current_user.id:
            include_personal_data = True
        elif current_user.is_admin:
            include_personal_data = True

    return {
        "status": "ok",
        "user": user.as_dict(include_personal_data=include_personal_data),
    }


@allow("authenticated", allow_blocked=True)
@schema("modify_user.json")
def post(user_id):
    """Modify an user"""
    if user_id != current_user.id:
        raise Forbidden("You can't modify this user")

    data = request.get_json()

    user = UserModel.get(id=user_id)

    password = data.get("password", None)
    token = data.get("token", None)

    if "new_password" in data or "email" in data:
        if not user.check_auth(password=password, token=token):
            raise Forbidden()

    if "new_password" in data:
        user.set_password(data["new_password"])

    if "email" in data:
        user.set_email(data["email"])
        user.send_email_change_mail()

    if "ui_preferences" in data:
        user.ui_preferences = json.dumps(data["ui_preferences"])

    current_api.database.session.commit()

    # personal data : user is current user, so always true
    return {"status": "ok", "user": user.as_dict(include_personal_data=True)}



import json
from tests.unit_tests.utils import BaseTest


class Test_Document(BaseTest):
    def assert_document(self, document, user, data, comment="creation"):
        assert document["comment"] == comment
        assert json.dumps(document["data"]) == json.dumps(data)
        assert isinstance(document["id"], int)
        assert isinstance(document["timestamp"], str)
        assert isinstance(document["version_id"], int)
        assert document["user"]["id"] == user.id

    def test_errors(self, user):
        self.create_document(expected_status=403)  # not logged

        self.login_user(user)

        fake_doc = {"id": 1, "version_id": 1, "data": None}

        self.get_document(fake_doc, expected_status=404)
        self.modify_document(fake_doc, expected_status=404)

        r = self.put("/documents", json={"comment": "xxx", "document": {"namespace": "x"}}, expected_status=400)
        assert r.json["description"] == "'data' is a required property on instance ['document']"

        r = self.put("/documents", json={"document": {"namespace": "x", "data": "x"}}, expected_status=400)
        assert r.json["description"] == "'comment' is a required property on instance "

    def test_creation(self, user):
        self.login_user(user)

        r = self.create_document(data={"value": "42"}, comment="XXX")
        self.assert_document(r.json["document"], user, data={"value": "42"}, comment="XXX")

        document_id = r.json["document"]["id"]

        r = self.get_document(document_id)
        self.assert_document(r.json["document"], user, data={"value": "42"}, comment="XXX")
        assert r.json["document"]["version_id"] == r.json["document"]["last_version_id"]

    def test_modification(self, user):
        self.login_user(user)

        v1 = self.create_document(expected_status=200).json["document"]
        v2 = self.modify_document(v1, comment="test", data={"value": "43"}, expected_status=200).json["document"]

        self.assert_document(v2, user, comment="test", data={"value": "43"})

        r = self.get_document(v1)
        assert r.json["document"]["version_id"] == r.json["document"]["last_version_id"]

        r = self.get_documents()
        assert r.json["status"] == "ok"
        assert r.json["count"] == 1
        assert r.json["documents"][0]["version_id"] == v2["version_id"]

    def test_deletion_error(self, admin):
        self.delete_document(1, expected_status=403)

        self.login_user(admin)

        self.delete_document(1, expected_status=404)

    def test_deletion(self, admin):
        self.login_user(admin)

        doc = self.create_document().json["document"]
        self.delete_document(doc, expected_status=200)
        self.get_document(doc, expected_status=404)
        self.get_version(doc, expected_status=404)

    def test_testing_helper(self, user):
        self.login_user(user)
        doc = self.create_document().json["document"]

        self.modify_document(doc, params={"rc_sleep": 0.01})



import json
from datetime import datetime

from flask_login import current_user
from sqlalchemy import Column, ForeignKey, Integer, String, DateTime, Boolean, select
from sqlalchemy.dialects.postgresql import ARRAY
from sqlalchemy.orm import relationship, reconstructor
from werkzeug.exceptions import BadRequest

from flask_camp._utils import current_api
from flask_camp.models._base_model import BaseModel
from flask_camp.models._user import User
from flask_camp.models._user_tag import UserTag


def _as_dict(document, version, include_hidden_data_for_staff=False):

    if document.is_redirection:
        return {"id": document.id, "redirect_to": document.redirect_to}

    result = {
        "id": document.id,
        "protected": document.protected,
        "comment": version.comment,
        "hidden": version.hidden,
        "timestamp": version.timestamp.isoformat(),
        "user": version.user.as_dict(),
        "last_version_id": document.last_version_id,
        "version_id": version.id,
    }

    if not version.hidden:
        result["data"] = version.data
    elif include_hidden_data_for_staff and (current_user.is_admin or current_user.is_moderator):
        result["data"] = version.data

    return result


class Document(BaseModel):
    __tablename__ = "document"

    id = Column(Integer, primary_key=True, index=True)

    protected = Column(Boolean, nullable=False, default=False)

    user_tags = relationship(UserTag, back_populates="document", lazy="select", cascade="all,delete")
    versions = relationship(
        lambda: DocumentVersion,
        primaryjoin=lambda: Document.id == DocumentVersion.document_id,
        backref="document",
        lazy="select",
        cascade="all,delete",
    )

    last_version_id = Column(Integer, ForeignKey("version.id", use_alter=True))
    last_version = relationship(
        lambda: DocumentVersion,
        primaryjoin=lambda: Document.last_version_id == DocumentVersion.id,
        uselist=False,
        post_update=True,
    )

    redirect_to = Column(Integer, ForeignKey("document.id"), index=True)

    associated_ids = Column(ARRAY(Integer), index=True)

    @classmethod
    def create(cls, comment, data, author=None):

        result = cls()

        version = DocumentVersion(
            document=result,
            user=current_user if author is None else author,
            comment=comment,
            data=data,
        )

        result.last_version = version
        result.associated_ids = current_api.get_associated_ids(version.as_dict())

        current_api.database.session.add(result)
        current_api.database.session.add(version)
        current_api.database.session.flush()

        return result

    def update_last_version_id(self, forbidden_id=None):
        """call this when a version has been hidden or deleted"""
        query = DocumentVersion.query

        if forbidden_id is not None:
            query = query.filter(DocumentVersion.id != forbidden_id)

        query = query.filter_by(document_id=self.id, hidden=False).order_by(DocumentVersion.id.desc())
        self.last_version = query.first()

        if self.last_version is None:
            raise BadRequest("There is no visible version associated with this document")

        self.last_version_id = self.last_version.id

    def clear_memory_cache(self):
        current_api.memory_cache.delete_document(self.id)

        query = select(Document.id).where(Document.associated_ids.contains([self.id]))
        for row in current_api.database.session.execute(query):
            current_api.memory_cache.delete_document(row[0])

    def as_dict(self):
        return _as_dict(self, self.last_version)

    @property
    def is_redirection(self):
        return self.redirect_to is not None


class DocumentVersion(BaseModel):
    __tablename__ = "version"

    document_id = Column(Integer, ForeignKey("document.id"), index=True)
    # document = relationship("Document", foreign_keys=[document_id], back_populates="versions")

    user_id = Column(Integer, ForeignKey(User.id), index=True)
    user = relationship(User)

    timestamp = Column(DateTime)
    comment = Column(String)

    hidden = Column(Boolean, default=False, nullable=False)
    _data = Column("data", String)

    user_tags = relationship(
        "UserTag",
        lazy="select",
        foreign_keys="DocumentVersion.document_id",
        primaryjoin=document_id == UserTag.document_id,
        uselist=True,
        viewonly=True,
    )

    def __init__(self, data, **kwargs):
        kwargs["timestamp"] = datetime.now()
        super().__init__(_data=json.dumps(data), **kwargs)
        self._init_from_database()

    @reconstructor
    def _init_from_database(self):
        self._raw_data = json.loads(self._data)

    def as_dict(self):
        return _as_dict(self.document, self, include_hidden_data_for_staff=True)

    @property
    def data(self):
        return self._raw_data

    # @data.setter
    # def data(self, value):
    #     self._raw_data = value
    #     self._data = json.dumps(value)



from flask import Flask

from flask_camp import RestApi
from flask_camp.client import ClientInterface


class BaseTest(ClientInterface):
    client = None
    rest_api_kwargs = {}

    @classmethod
    def setup_class(cls):
        cls.app = Flask(__name__, static_folder=None)
        cls.app.config.update(
            {"TESTING": True, "SECRET_KEY": "not very secret", "SQLALCHEMY_TRACK_MODIFICATIONS": False}
        )

        cls.api = RestApi(app=cls.app, **cls.rest_api_kwargs)

    def setup_method(self):
        with self.app.app_context():
            self.api.database.create_all()

        with self.app.test_client() as client:
            self.client = client

    def teardown_method(self):
        with self.app.app_context():
            self.api.database.drop_all()

        self.api.memory_cache.flushall()

    @staticmethod
    def _convert_kwargs(kwargs):
        """convert request argument to flask test client argument"""
        kwargs["query_string"] = kwargs.pop("params", None)

    def get(self, url, **kwargs):
        expected_status = kwargs.pop("expected_status", 200)
        self._convert_kwargs(kwargs)

        r = self.client.get(url, **kwargs)
        self.assert_status_code(r, expected_status)

        return r

    def post(self, url, **kwargs):
        expected_status = kwargs.pop("expected_status", 200)
        self._convert_kwargs(kwargs)

        r = self.client.post(url, **kwargs)
        self.assert_status_code(r, expected_status)

        return r

    def put(self, url, **kwargs):
        expected_status = kwargs.pop("expected_status", 200)
        self._convert_kwargs(kwargs)

        r = self.client.put(url, **kwargs)
        self.assert_status_code(r, expected_status)

        return r

    def delete(self, url, **kwargs):
        expected_status = kwargs.pop("expected_status", 200)
        self._convert_kwargs(kwargs)

        r = self.client.delete(url, **kwargs)
        self.assert_status_code(r, expected_status)

        return r

    @staticmethod
    def assert_status_code(response, expected_status):
        if expected_status is None:
            expected_status = [200]
        elif isinstance(expected_status, int):
            expected_status = [expected_status]

        assert (
            response.status_code in expected_status
        ), f"Status error: {response.status_code} i/o {expected_status}\n{response.data}"

    def _assert_status_response(self, r):
        if r.status_code == 304:  # not modified : no body
            assert r.data == b""
        elif r.status_code == 301:  # Moved permanently : no body
            assert r.data == b""
            assert "Location" in r.headers
        else:
            assert r.json is not None, r.data
            assert "status" in r.json, r.json

            if r.status_code == 200:
                assert r.json["status"] == "ok", r.json
            else:
                assert r.json["status"] == "error", r.json
                assert "description" in r.json, r.json

    def create_user(self, name="user", email=None, password="password", **kwargs):
        email = f"{name}@example.com" if email is None else email
        return super().create_user(name, email, password, **kwargs)

    def login_user(self, user, password=None, token=None, **kwargs):
        if password is None and token is None:
            password = "password"

        return super().login_user(user, password=password, token=token, **kwargs)

    def block_user(self, user, comment="Default comment", **kwargs):
        return super().block_user(user, comment, **kwargs)

    def unblock_user(self, user, comment="Default comment", **kwargs):
        return super().unblock_user(user, comment, **kwargs)

    def create_document(self, data=None, comment="Default comment", **kwargs):
        return super().create_document(data={} if data is None else data, comment=comment, **kwargs)

    def modify_document(self, document, comment="Default comment", data=None, **kwargs):
        data = data if data is not None else document["data"]
        return super().modify_document(document, comment, data, **kwargs)

    def get_document(
        self,
        document,
        data_should_be_present=True,
        version_should_be=None,
        **kwargs,
    ):
        r = super().get_document(document, **kwargs)

        if r.status_code == 200:
            if data_should_be_present:
                assert "data" in r.json["document"]
            else:
                assert "data" not in r.json["document"]

            if version_should_be:
                assert r.json["document"]["version_id"] == version_should_be["version_id"], r.json["document"]

        return r

    def protect_document(self, document, comment="Default comment", **kwargs):
        return super().protect_document(document, comment, **kwargs)

    def unprotect_document(self, document, comment="Default comment", **kwargs):
        return super().unprotect_document(document, comment, **kwargs)

    def merge_documents(self, document_to_merge, document_destination, comment="Default comment", **kwargs):
        return super().merge_documents(document_to_merge, document_destination, comment, **kwargs)

    def delete_document(self, document, comment="Default comment", **kwargs):
        return super().delete_document(document, comment, **kwargs)

    def get_version(self, version, data_should_be_present=True, **kwargs):
        r = super().get_version(version, **kwargs)

        if r.status_code == 200:
            if data_should_be_present:
                assert "data" in r.json["document"]
            else:
                assert "data" not in r.json["document"]

        return r

    def hide_version(self, version, comment="Default comment", **kwargs):
        return super().hide_version(version, comment, **kwargs)

    def unhide_version(self, version, comment="Default comment", **kwargs):
        return super().unhide_version(version, comment, **kwargs)

    def delete_version(self, version, comment="Default comment", **kwargs):
        return super().delete_version(version, comment, **kwargs)

    def rename_user(self, user, name, comment="Default comment", **kwargs):
        return super().rename_user(user, name, comment, **kwargs)



from flask import request
from werkzeug.exceptions import BadRequest

from flask_camp._utils import current_api
from flask_camp._services._security import allow
from flask_camp.models._document import DocumentVersion
from flask_camp.models._user import User
from flask_camp._utils import cook

rule = "/versions"


@allow("anonymous", "authenticated", allow_blocked=True)
def get():
    """Get a list of versions"""
    filters = {}

    limit = request.args.get("limit", default=30, type=int)
    offset = request.args.get("offset", default=0, type=int)
    document_id = request.args.get("document_id", default=None, type=int)
    user_id = request.args.get("user_id", default=None, type=str)

    if not 0 <= limit <= 100:
        raise BadRequest("Limit can't be lower than 0 or higher than 100")

    tag_filters_args = {
        "user_id": request.args.get("tag_user_id", default=None, type=int),
        "name": request.args.get("tag_name", default=None, type=str),
        "value": request.args.get("tag_value", default=None, type=str),
    }

    tag_filters_args = {k: v for k, v in tag_filters_args.items() if v is not None}

    query = current_api.database.session.query(DocumentVersion)

    if len(tag_filters_args) != 0:
        query = query.filter(DocumentVersion.user_tags.any(**tag_filters_args))

    if document_id is not None:
        filters["document_id"] = document_id

    if len(filters) != 0:
        query = query.filter_by(**filters)

    if user_id is not None:
        query = query.join(User).filter(User.id == user_id)

    query = query.order_by(DocumentVersion.id.desc())
    count = query.count()
    versions = query.offset(offset).limit(limit)

    return {
        "status": "ok",
        "count": count,
        "versions": [cook(version.as_dict()) for version in versions],
    }



from freezegun import freeze_time
from flask_login import current_user

from flask_camp import allow

from tests.unit_tests.utils import BaseTest


class RateLimitedClass:
    rule = "/rate_limited"

    @allow("anonymous")
    def get(self):
        """I'm rate limited"""
        return {"hello": "world"}


class Test_RateLimit(BaseTest):
    rest_api_kwargs = {
        "rate_limit_cost_function": lambda: 0 if current_user.is_admin else 1,
        "rate_limits_file": "tests/ratelimit_config.json",
    }

    def test_rate_limited_class_instance(self):

        self.api.add_modules(self.app, RateLimitedClass())

        results = []
        with freeze_time():
            for _ in range(6):
                r = self.get("/rate_limited", expected_status=(200, 429))
                results.append(r.status_code)

        assert results[0] == results[1] == results[2] == results[3] == results[4] == 200
        assert results[5] == 429

    def test_main(self):
        results = []
        with freeze_time():
            for _ in range(6):
                r = self.get("/", environ_base={"REMOTE_ADDR": "127.0.0.2"}, expected_status=(200, 429))
                results.append(r.status_code)

        assert results[0] == results[1] == results[2] == results[3] == results[4] == 200
        assert results[5] == 429

    def test_rate_limit_cost(self, admin):

        results = []

        self.login_user(admin)
        with freeze_time():
            for _ in range(6):
                r = self.get("/", environ_base={"REMOTE_ADDR": "127.0.0.2"}, expected_status=(200, 429))
                results.append(r.status_code)

        assert results[0] == results[1] == results[2] == results[3] == results[4] == results[5] == 200

    def test_info(self):
        routes = self.get("/").json
        assert routes["/"]["GET"].get("rate_limit") == "5 per second", routes["/"]["GET"]



from flask import request
from werkzeug.exceptions import NotFound, BadRequest

from flask_camp._schemas import schema
from flask_camp._utils import current_api
from flask_camp.models._document import Document, DocumentVersion
from flask_camp._services._security import allow

rule = "/merge"


@allow("moderator")
@schema("merge_documents.json")
def post():
    """Merge two documents. Merged document will become a redirection, and will be no longer modifiable
    Other document will get all hostory from merged"""

    data = request.get_json()
    document_to_merge = Document.get(id=data["document_to_merge"], with_for_update=True)
    document_destination = Document.get(id=data["document_destination"], with_for_update=True)

    if document_to_merge is None or document_destination is None:
        raise NotFound()

    if document_to_merge.id == document_destination.id:
        raise BadRequest()

    if document_destination.is_redirection or document_to_merge.is_redirection:
        raise BadRequest()

    document_to_merge.redirect_to = document_destination.id
    DocumentVersion.query.filter_by(document_id=document_to_merge.id).update({"document_id": document_destination.id})
    document_to_merge.last_version_id = None
    document_to_merge.last_version = None
    document_destination.update_last_version_id()

    current_api.before_document_save(document_to_merge)
    current_api.before_document_save(document_destination)
    current_api.add_log(
        "merge", comment=data["comment"], document=document_destination, merged_document=document_to_merge
    )
    current_api.database.session.commit()

    document_destination.clear_memory_cache()
    document_to_merge.clear_memory_cache()

    return {"status": "ok"}



from flask import request
from werkzeug.exceptions import NotFound, BadRequest

from flask_camp._schemas import schema
from flask_camp._utils import current_api
from flask_camp.models._user import User as UserModel
from flask_camp._services._security import allow

rule = "/block_user/<int:user_id>"


@allow("moderator")
@schema("action_with_comment.json")
def post(user_id):
    """Block/unblock an user"""

    user = UserModel.get(id=user_id, with_for_update=True)

    if not user:
        raise NotFound()

    blocked = request.get_json()["blocked"]

    if blocked == user.blocked:
        raise BadRequest("User is still blocked/unblocked")

    user.blocked = blocked

    current_api.add_log(action="block" if blocked else "unblock", target_user=user)
    current_api.database.session.commit()

    return {"status": "ok"}



from flask import request, current_app
from flask_login import current_user
from sqlalchemy.exc import IntegrityError
from werkzeug.exceptions import BadRequest

from flask_camp._schemas import schema
from flask_camp._utils import current_api
from flask_camp._services._security import allow
from flask_camp.models._user import User as UserModel

rule = "/users"


@allow("moderator")
def get():
    """Get a list of users"""

    limit = request.args.get("limit", default=30, type=int)
    offset = request.args.get("offset", default=0, type=int)

    if not 0 <= limit <= 100:
        raise BadRequest("Limit can't be lower than 0 or higher than 100")

    query = UserModel.query

    users = query.order_by(UserModel.id.desc()).limit(limit).offset(offset)

    return {"status": "ok", "users": [user.as_dict() for user in users], "count": query.count()}


@allow("anonymous", "authenticated")
@schema("create_user.json")
def put():
    """create an user"""

    if current_user.is_authenticated:
        raise BadRequest()

    data = request.get_json()

    user = UserModel(name=data["name"])
    user.set_password(data["password"])
    user.set_email(data["email"])

    current_api.database.session.add(user)

    try:
        current_api.database.session.flush()
        current_api.before_user_creation(user)
        current_api.database.session.commit()
    except IntegrityError as e:
        raise BadRequest("A user still exists with this name") from e

    try:
        user.send_account_creation_mail()
    except:  # pylint: disable=bare-except
        current_app.logger.exception("Fail to send mail", exc_info=True)

    return {"status": "ok", "user": user.as_dict()}



import collections
from time import perf_counter

import requests
from sqlalchemy import create_engine

from flask_camp.client import ClientInterface


engine = create_engine("postgresql://flask_camp_user:flask_camp_user@localhost:5432/flask_camp")


def get_email_token(user_name):
    with engine.connect() as connection:
        rows = connection.execute(f"SELECT email_token FROM user_account WHERE name='{user_name}'")
        token = list(rows)[0][0]
        return token


class ClientSession(ClientInterface):
    _session_id = 0

    def __init__(self, domain="http://localhost:5000"):
        self.domain = domain
        self._session = requests.Session()
        self.logged_user = None
        ClientSession._session_id += 1  # TODO thread lock
        self.session_id = ClientSession._session_id
        self.stats = collections.defaultdict(list)

    @property
    def logged_user_name(self):
        return "<anon>" if self.logged_user is None else self.logged_user["name"]

    def __str__(self):
        return f"S#{self.session_id} ({self.logged_user_name})"

    def add_stat(self, method, url, outcome, ellapsed):
        key = (method.upper(), url.split("/")[1], outcome)
        self.stats[key].append(ellapsed)

    def _request(self, method, url, **kwargs):
        outcome = None

        expected_status = kwargs.pop("expected_status", None)

        start = perf_counter()
        try:
            r = self._session.request(method, f"{self.domain}{url}", **kwargs, timeout=10)
            outcome = str(r.status_code)
        except Exception as e:
            outcome = str(e)
            raise
        finally:
            ellapsed = (perf_counter() - start) * 1000
            self.add_stat(method, url, outcome, ellapsed)

        expected_status = 200 if expected_status is None else expected_status

        expected_status = expected_status if isinstance(expected_status, (list, tuple, set)) else [expected_status]

        if r.status_code not in expected_status:
            print(f"{method.upper()} {url}")
            print(f"Params: {kwargs.get('params')}")
            print(f"Expected status: {expected_status}")
            print(f"Observed status: {r.status_code}")
            print(f"Headers: {r.headers}")
            print(f"Content: {r.content}")
            raise ValueError()

        return r

    def get(self, url, **kwargs):
        return self._request("get", url, **kwargs)

    def post(self, url, **kwargs):
        return self._request("post", url, **kwargs)

    def put(self, url, **kwargs):
        return self._request("put", url, **kwargs)

    def delete(self, url, **kwargs):
        return self._request("delete", url, **kwargs)

    def login_user(self, user, password="password", token=None, **kwargs):
        r = super().login_user(user, password=password, token=token, **kwargs)

        if r.status_code == 200:
            self.logged_user = r.json()["user"]

        return r

    def logout_user(self, **kwargs):
        r = super().logout_user(**kwargs)

        if r.status_code == 200:
            self.logged_user = None

        return r

    def setup_user(self, name):
        self.create_user(name, email=f"{name}@example.com", password="password")
        self.validate_email(name, get_email_token(name))
        self.login_user(name)

    @property
    def is_anonymous(self):
        return self.logged_user is None

    @property
    def is_moderator(self):
        return not self.is_anonymous and "moderator" in self.logged_user["roles"]



import pytest

from flask_camp import RestApi
from flask_camp.exceptions import ConfigurationError

from tests.unit_tests.utils import BaseTest


def cooker(document, get_document):
    data = document.get("data")

    if isinstance(data, dict) and data.get("namespace") in ("cook-me",):
        document["cooked"] = {}

        # Let's build an app with document. One rule: all documents have (or not) a parent
        # if a document has a parent, it must be present in document["cooked"]["parent"]
        parent_id = document["data"].get("parent_id")
        if parent_id is not None:
            document["cooked"]["parent"] = get_document(parent_id)
        else:
            document["cooked"]["parent"] = None


class Test_Cooker(BaseTest):
    rest_api_kwargs = {"cooker": cooker}

    def test_error(self):
        with pytest.raises(ConfigurationError):
            RestApi(cooker={})

    def test_basic(self, user):
        self.login_user(user)

        doc = self.create_document(data={"namespace": "cook-me", "value": "42"}).json["document"]
        assert "cooked" in doc, doc

        doc = self.get_document(doc).json["document"]
        assert "cooked" in doc

        doc = self.modify_document(doc).json["document"]
        assert "cooked" in doc

        doc = self.get_version(doc).json["document"]
        assert "cooked" in doc

        docs = self.get_documents().json["documents"]
        assert "cooked" in docs[0]

        docs = self.get_versions().json["versions"]
        assert "cooked" in docs[0]

        self.add_user_tag("star", doc)
        docs = self.get_documents(tag_name="star").json["documents"]
        assert "cooked" in docs[0]

        assert "cooked" not in self.api.memory_cache.get_document(doc["id"])
        assert "cooked" in self.api.memory_cache.get_cooked_document(doc["id"])

    def test_association(self, moderator, admin):
        # Reminder: the tested app havea cooker: documents have (or not) a parent
        # if a document has a parent, it must be present in document["parent"]

        self.login_user(moderator)

        # create a doc
        parent_v1 = self.create_document(
            data={"namespace": "cook-me", "parent_id": None, "content": "v1"},
        ).json["document"]

        # create another doc. It's parent is doc_1
        child = self.create_document(data={"namespace": "cook-me", "parent_id": parent_v1["id"]}).json["document"]
        assert "cooked" in child
        assert "parent" in child["cooked"]
        assert child["cooked"]["parent"]["id"] == parent_v1["id"]
        assert child["cooked"]["parent"]["data"]["content"] == "v1"

        # now, modify the parent
        parent_v2 = self.modify_document(
            parent_v1, data={"namespace": "cook-me", "parent_id": None, "content": "v2"}
        ).json["document"]

        # the child must have the updated value of the parent
        child = self.get_document(child).json["document"]

        # not let's try to hide a version, the child must have the first value of the parent
        self.hide_version(parent_v2)
        child = self.get_document(child).json["document"]
        assert child["cooked"]["parent"]["data"]["content"] == "v1"
        assert child["cooked"]["parent"]["version_id"] == parent_v1["version_id"]

        # unhide it, back to parent v2
        self.unhide_version(parent_v2)
        child = self.get_document(child).json["document"]
        assert child["cooked"]["parent"]["data"]["content"] == "v2"
        assert child["cooked"]["parent"]["version_id"] == parent_v2["version_id"]

        # delete parent v2, again v1 should appear
        self.login_user(admin)
        self.delete_version(parent_v2)
        child = self.get_document(child).json["document"]
        assert child["cooked"]["parent"]["data"]["content"] == "v1"
        assert child["cooked"]["parent"]["version_id"] == parent_v1["version_id"]

        # Now, delete the parent
        self.delete_document(parent_v1)
        child = self.get_document(child).json["document"]
        assert child["cooked"]["parent"] is None

    def test_circular_reference(self, user):

        self.login_user(user)
        twin_a = self.create_document(data={"namespace": "cook-me", "parent_id": None}).json["document"]
        twin_b = self.create_document(data={"namespace": "cook-me", "parent_id": twin_a["id"]}).json["document"]

        twin_a = self.modify_document(twin_a, data={"namespace": "cook-me", "parent_id": twin_b["id"]}).json["document"]
        assert twin_a["cooked"]["parent"]["version_id"] == twin_b["version_id"]

        twin_a = self.get_document(twin_a).json["document"]
        assert twin_a["cooked"]["parent"]["version_id"] == twin_b["version_id"]

        twin_b = self.get_document(twin_b).json["document"]
        assert twin_b["cooked"]["parent"]["version_id"] == twin_a["version_id"]

        # self reference
        narcissus = self.create_document(data={"namespace": "cook-me", "parent_id": None}).json["document"]
        narcissus = self.modify_document(narcissus, data={"namespace": "cook-me", "parent_id": narcissus["id"]}).json[
            "document"
        ]

        assert "cooked" in narcissus, narcissus
        assert "parent" in narcissus["cooked"], narcissus["cooked"]
        assert narcissus["cooked"]["parent"]["version_id"] == narcissus["version_id"]



from redis import Redis as RedisClient
from redis.commands.json.path import Path


class _MemoryCacheCollection:
    def __init__(self, name, client):
        self.name = name
        self._client = client.json()

    def get(self, document_id):
        return self._client.get(f"{self.name}:{document_id}")

    def set(self, id, document):
        self._client.set(f"{self.name}:{id}", Path.root_path(), document)

    def delete(self, id):
        self._client.delete(f"{self.name}:{id}")


class _MemoryCache:
    def __init__(self):
        self._client = None
        self._document = None

    def init_app(self, app):
        host = app.config.get("REDIS_HOST", "localhost")
        port = app.config.get("REDIS_PORT", 6379)

        self._client = RedisClient(host=host, port=port)
        self._document = _MemoryCacheCollection("document", self._client)

    def set_document(self, document_id, document_as_dict, cooked_document_as_dict):
        self._document.set(document_id, {"document": document_as_dict, "cooked_document": cooked_document_as_dict})

    def get_document(self, document_id):
        result = self._document.get(document_id)

        return None if result is None else result["document"]

    def get_cooked_document(self, document_id):
        result = self._document.get(document_id)

        return None if result is None else result["cooked_document"]

    def delete_document(self, document_id):
        self._document.delete(document_id)

    def flushall(self):
        self._client.flushall()


memory_cache = _MemoryCache()



from functools import wraps

from flask_login import current_user
from werkzeug.exceptions import Forbidden


def allow(*args, allow_blocked=False):
    """simple decorator that add roles attributes to a function"""

    allowed_roles = set(args)

    def decorator(f):

        setattr(f, "allowed_roles", allowed_roles)
        setattr(f, "allow_blocked", allow_blocked)

        return f

    return decorator


def check_rights(f):
    """Takes a function in argument, and return it decorated with a right checker
    The function must have been decorated woth the @allow decorator"""

    @wraps(f)
    def wrapper(*args, **kwargs):

        if current_user.blocked and not f.allow_blocked:
            raise Forbidden("You have been blocked, you can't access to this resource")

        user_roles = list(current_user.roles)

        if current_user.is_authenticated:
            user_roles.append("authenticated")

        for user_role in user_roles:
            if user_role in f.allowed_roles:
                return f(*args, **kwargs)

        raise Forbidden("You can't access to this resource")

    return wrapper



import copy
import json
from types import ModuleType
import warnings

from flask import request
from flask_login import LoginManager
from flask_limiter import Limiter
from flask_limiter.util import get_remote_address
from werkzeug.exceptions import HTTPException, NotFound

from ._schemas import SchemaValidator
from ._services._database import database
from ._services._memory_cache import memory_cache
from ._services._security import check_rights, allow
from ._services._send_mail import SendMail
from ._utils import GetDocument
from .exceptions import ConfigurationError
from .models._document import Document
from .models._log import Log
from .models._user import User as UserModel, AnonymousUser


from .views.account import block_user as block_user_view
from .views.account import current_user as current_user_view
from .views.account import email_validation as email_validation_view
from .views.account import rename_user as rename_user_view
from .views.account import reset_password as reset_password_view
from .views.account import roles as roles_view
from .views.account import user as user_view
from .views.account import user_login as user_login_view
from .views.account import users as users_view
from .views.content import document as document_view
from .views.content import documents as documents_view
from .views.content import version as version_view
from .views.content import versions as versions_view
from .views.content import merge as merge_view
from .views.content import protect_document as protect_document_view
from .views.content import user_tags as user_tags_view
from .views import healthcheck as healthcheck_view
from .views import home as home_view
from .views import logs as logs_view


# pylint: disable=too-many-instance-attributes
class RestApi:
    def __init__(  # pylint: disable=too-many-locals
        self,
        app=None,
        cooker=None,
        schemas_directory=None,
        document_schemas=None,
        user_roles="",
        rate_limit_cost_function=None,
        rate_limits_file=None,
        user_can_delete=False,
        before_user_creation=None,
        before_document_save=None,
        before_document_delete=None,
        update_search_query=None,
    ):
        self.database = database
        self.limiter = Limiter(key_func=get_remote_address)
        self.memory_cache = memory_cache
        self.mail = SendMail()

        self._rate_limit_cost_function = rate_limit_cost_function

        self.user_can_delete = user_can_delete

        self.before_user_creation = self._hook_function(before_user_creation)
        self.before_document_save = self._hook_function(before_document_save)
        self.before_document_delete = self._hook_function(before_document_delete)
        self.update_search_query = update_search_query if update_search_query is not None else lambda query: query

        if rate_limits_file:
            with open(rate_limits_file, mode="r", encoding="utf-8") as f:
                self._rate_limits = json.load(f)
        else:
            self._rate_limits = {}

        self._user_roles = {"admin", "moderator"} | self._parse_user_roles(user_roles)
        self._cooker = cooker

        if schemas_directory:
            self._schema_validator = SchemaValidator(schemas_directory)
        else:
            self._schema_validator = None

        self._document_schemas = document_schemas
        self.allow = allow

        if app is not None:
            self.init_app(app)
        else:
            self._configuration_checks()

    def init_app(self, app):
        app.extensions["flask-camp"] = self

        self._init_config(app)
        self._init_database(app)
        self.memory_cache.init_app(app)
        self.mail.init_app(app)
        self._init_login_manager(app)
        self._init_error_handler(app)
        self._init_rate_limiter(app)
        self._init_url_rules(app)

        self._configuration_checks()

        if app.debug:  # pragma: no cover
            with app.app_context():
                self.database.create_all()

    def _configuration_checks(self):
        # post configuration checks
        if self._cooker is not None and not callable(self._cooker):
            raise ConfigurationError(f"cooker is not callable: {self._cooker}")

        if self._schema_validator and self._document_schemas:
            for filename in self._document_schemas:
                if not self._schema_validator.exists(filename):
                    raise FileNotFoundError(f"Schema {filename} does not exists")

        for role in ("anonymous", "authenticated"):
            if role in self._user_roles:
                raise ConfigurationError(f"{role} can't be a user role")

    @staticmethod
    def _parse_user_roles(user_roles):
        if isinstance(user_roles, str):  # allow comma separated string
            user_roles = user_roles.split(",")

        return set(role.lower().strip() for role in user_roles if len(role.strip()) != 0)

    @staticmethod
    def _hook_function(function):

        if function is not None and not callable(function):
            raise ConfigurationError(f"Hook object is not callable: {function}")

        return function if function else lambda *args, **kwargs: ...

    def _init_config(self, app):
        if app.config.get("MAIL_DEFAULT_SENDER", None) is None:
            if not app.testing and not app.debug:
                warnings.warn("MAIL_DEFAULT_SENDER is not set, defaulting to do-not-reply@example.com")
            app.config["MAIL_DEFAULT_SENDER"] = "do-not-reply@example.com"

    def _init_database(self, app):

        if app.config.get("SQLALCHEMY_DATABASE_URI", None) is None:
            default = "postgresql://flask_camp_user:flask_camp_user@localhost:5432/flask_camp"
            app.config["SQLALCHEMY_DATABASE_URI"] = default

            if not app.testing and not app.debug:
                warnings.warn(f"SQLALCHEMY_DATABASE_URI is not set, defaulting to {default}")

        self.database.init_app(app)

        @app.teardown_appcontext
        def shutdown_session(exception=None):  # pylint: disable=unused-argument
            self.database.session.remove()

    def _init_login_manager(self, app):
        login_manager = LoginManager(app)
        login_manager.anonymous_user = AnonymousUser

        @login_manager.user_loader  # pylint: disable=no-member
        def load_user(user_id):
            return UserModel.get(id=int(user_id))

    def _init_error_handler(self, app):
        @app.errorhandler(HTTPException)
        def rest_error_handler(e):
            result = {"status": "error", "name": e.name, "description": e.description}
            if hasattr(e, "data"):
                result["data"] = e.data
            return result, e.code

    def _init_rate_limiter(self, app):

        if "RATELIMIT_STORAGE_URI" not in app.config:
            redis_host = app.config.get("REDIS_HOST", "localhost")
            redis_port = app.config.get("REDIS_PORT", 6379)
            app.config["RATELIMIT_STORAGE_URI"] = f"redis://{redis_host}:{redis_port}"

        self.limiter.init_app(app)

    def _init_url_rules(self, app):
        # basic page: home and healtcheck
        self.add_modules(app, home_view, healthcheck_view)

        # access to users
        self.add_modules(app, users_view, user_view, current_user_view)

        # related to user account
        self.add_modules(app, user_login_view, email_validation_view, reset_password_view)

        # related to document
        self.add_modules(app, documents_view, document_view)
        self.add_modules(app, versions_view, version_view)
        self.add_modules(app, user_tags_view)

        # logs
        self.add_modules(app, logs_view)

        # reserved for moderators
        self.add_modules(app, protect_document_view)
        self.add_modules(app, block_user_view, rename_user_view)
        self.add_modules(app, merge_view)

        # reserved for admins
        self.add_modules(app, roles_view)

    @property
    def user_roles(self):
        return set(self._user_roles)

    ### Public methods

    def get_associated_ids(self, document_as_dict):
        associated_ids = []

        if self._cooker is not None:
            get_document = GetDocument(self.get_document)
            self._cooker(copy.deepcopy(document_as_dict), get_document)
            associated_ids = list(get_document.loaded_document_ids)

        return associated_ids

    def get_document(self, document_id):
        """This very simple function get a document id and returns it as a dict.
        It's only puprose it to hide the memcache complexity"""
        document_as_dict = memory_cache.get_document(document_id)

        if document_as_dict is None:  # document is not known by mem cache
            document = Document.get(id=document_id)

            if document is None:
                raise NotFound()

            document_as_dict = document.as_dict()

        return document_as_dict

    def get_cooked_document(self, document_id):
        """This very simple function get a document id and returns it as a dict.
        It's only puprose it to hide the memcache complexity"""
        cooked_document_as_dict = memory_cache.get_cooked_document(document_id)

        if cooked_document_as_dict is None:  # document is not known by mem cache
            document = Document.get(id=document_id)

            if document is None:
                raise NotFound()

            document_as_dict = document.as_dict()
            if document.is_redirection:
                cooked_document_as_dict = document_as_dict
            else:
                cooked_document_as_dict = self.cook(document_as_dict, save_in_memory_cache=True)

        return cooked_document_as_dict

    def cook(self, document_as_dict, save_in_memory_cache=False):
        result = copy.deepcopy(document_as_dict)

        if self._cooker is not None:
            self._cooker(result, GetDocument(self.get_document))

        if save_in_memory_cache:
            memory_cache.set_document(document_as_dict["id"], document_as_dict, result)

        return result

    def validate_user_schemas(self, data):
        if self._schema_validator is not None:
            self._schema_validator.validate(data, *self._document_schemas)

    def add_modules(self, app, *modules):
        possible_user_roles = self.user_roles | {"anonymous", "authenticated"}

        for module in modules:
            if not hasattr(module, "rule"):
                raise ConfigurationError(f"{module} does not have a rule attribute")

            for method in ["get", "post", "put", "delete"]:
                if hasattr(module, method):
                    function = getattr(module, method)
                    method = method.upper()

                    if not hasattr(function, "allowed_roles") or not hasattr(function, "allow_blocked"):
                        raise ConfigurationError(f"Please use @flask_camp.allow decorator on {function}")

                    for role in function.allowed_roles:
                        if role not in possible_user_roles:
                            raise ConfigurationError(f"{role} is not recognised")

                    function = check_rights(function)

                    if module.rule in self._rate_limits and method in self._rate_limits[module.rule]:
                        limit = self._rate_limits[module.rule][method]
                        if limit is not None:
                            function = self.limiter.limit(limit, cost=self._rate_limit_cost_function)(function)
                            app.logger.info("Use %s rate limit for %s %s", limit, method, module.rule)
                        else:
                            function = self.limiter.exempt(function)

                    if isinstance(module, ModuleType):
                        endpoint = f"{method}_{module.__name__}_{module.rule}"
                    else:
                        endpoint = f"{method}_{module.__class__.__name__}_{module.rule}"

                    app.add_url_rule(
                        module.rule,
                        view_func=function,
                        methods=[method],
                        endpoint=endpoint,
                    )

    def add_log(self, action, comment=None, target_user=None, document=None, merged_document=None, version=None):
        if comment is None:
            comment = request.get_json().get("comment", "")

        document_id = document.id if document is not None else None
        merged_document_id = merged_document.id if merged_document is not None else None
        version_id = version.id if version is not None else None

        log = Log(
            action=action,
            comment=comment,
            target_user=target_user,
            document_id=document_id,
            merged_document_id=merged_document_id,
            version_id=version_id,
        )
        self.database.session.add(log)



from tests.unit_tests.utils import BaseTest


class Test_Home(BaseTest):
    def test_main(self):
        response = self.get("/")
        assert response.status_code == 200

        for entry_point in response.json:
            for method in response.json[entry_point]:
                data = response.json[entry_point][method]
                assert data["description"] is not None, f"Docstring is not set for {method} {entry_point}"

                if not entry_point.startswith("/__"):
                    assert "allowed_roles" in data is not None, f"allowed_roles is not set for {method} {entry_point}"
                    assert "allow_blocked" in data is not None, f"allow_blocked is not set for {method} {entry_point}"



from flask import request
from werkzeug.exceptions import NotFound, BadRequest

from flask_camp._schemas import schema
from flask_camp._utils import current_api
from flask_camp.models._document import Document
from flask_camp._services._security import allow

rule = "/protect_document/<int:document_id>"


@allow("moderator")
@schema("protect_document.json")
def post(document_id):
    """Protect/unprotect a document. The document won't be editable anymore, except for moderators"""
    document = Document.get(id=document_id, with_for_update=True)

    if document is None:
        raise NotFound()

    if document.is_redirection:
        raise BadRequest()

    protected = request.get_json()["protected"]

    if protected == document.protected:
        raise BadRequest("User is still blocked/unblocked")

    document.protected = protected

    current_api.add_log("protect" if protected else "unprotect", document=document)
    current_api.database.session.commit()

    document.clear_memory_cache()

    return {"status": "ok"}



from flask import current_app
from werkzeug.exceptions import NotFound
from werkzeug.local import LocalProxy


current_api = LocalProxy(lambda: current_app.extensions["flask-camp"])
get_cooked_document = LocalProxy(lambda: current_api.get_cooked_document)
get_document = LocalProxy(lambda: current_api.get_document)
cook = LocalProxy(lambda: current_api.cook)


class GetDocument:  # pylint: disable=too-few-public-methods
    """This class is a callable the memorize wich arguments has been called
    It's used for the cooker
    """

    def __init__(self, original_get_document):
        self.loaded_document_ids = set()
        self.original_get_document = original_get_document

    def __call__(self, document_id):
        self.loaded_document_ids.add(document_id)
        try:
            return self.original_get_document(document_id)
        except NotFound:
            # it's a possible outcome, if the document has been deleted
            # In that situation, returns None
            return None



from tests.unit_tests.utils import BaseTest


# TODO : test block when user is currently logged in


class Test_UserBlock(BaseTest):
    def test_not_allowed(self, user):
        self.block_user(user, expected_status=403)
        self.unblock_user(user, expected_status=403)

        self.login_user(user)

        self.block_user(user, expected_status=403)
        self.unblock_user(user, expected_status=403)

    def test_not_found(self, moderator):
        self.login_user(moderator)

        self.block_user(42, expected_status=404)
        self.unblock_user(42, expected_status=404)

    def test_typical_scenario(self, moderator, user):
        # log moderator, create a doc
        self.login_user(moderator)

        doc = self.create_document().json["document"]

        # now get the user, check its blocked status, and block him
        r = self.get_user(user)
        assert r.json["user"]["blocked"] is False

        self.block_user(user)
        self.block_user(user, expected_status=400)  # block him twice, it should produce an bad request error

        r = self.get_user(user)  # it's status is now blocked
        assert r.json["user"]["blocked"] is True

        self.logout_user()

        # user login and try to get/add/modify a doc
        self.login_user(user, expected_status=200)
        self.get_document(doc, expected_status=200)
        self.get_documents(expected_status=200)

        self.create_document(expected_status=403)
        self.modify_document(doc, expected_status=403)

        # Though, he can modify itself
        self.modify_user(user, new_password="updated", password="password", expected_status=200)

        # even get users, or one user
        self.get_user(moderator, expected_status=200)

        # logout the user, login the moderator, unblock the user
        self.logout_user()
        self.login_user(moderator)

        self.unblock_user(user)
        self.unblock_user(user, expected_status=400)  # unblock him twice, it should produce a bad request error

        r = self.get_user(user)
        assert not r.json["user"]["blocked"]

        # logout the admin, login the user, try to add/modify
        self.logout_user()
        self.login_user(user, password="updated")

        self.create_document(expected_status=200)
        self.modify_document(doc, data={"value": "42"}, expected_status=200)



import logging

from flask import Flask
from flask_camp import RestApi


app = Flask(__name__, static_folder=None)
app.config.update({"SECRET_KEY": "not very secret", "SQLALCHEMY_TRACK_MODIFICATIONS": False})
app.config.from_prefixed_env()

handler = logging.FileHandler("logs/testing_errors.log")
handler.setLevel(logging.ERROR)
app.logger.addHandler(handler)  # pylint: disable=no-member

api = RestApi(app=app)

if __name__ == "__main__":
    with app.app_context():
        api.database.create_all()



import collections
from datetime import datetime, timedelta
import json
import random
import string
import threading
import time

import docker

from tests.end_tests.utils import ClientSession


class FuzzerSession(ClientSession):
    def __init__(self, session_count):
        super().__init__()
        self.known_documents = None
        self.kill = False
        self.session_count = session_count

    def fuzz_get_user(self):
        self.get_users(expected_status=[200, 403])

    def fuzz_block_user(self):
        user_id = random.randint(4, self.session_count)
        while user_id == self.logged_user["id"]:
            user_id = random.randint(1, self.session_count)

        self.block_user(user_id, comment="Ha!", expected_status=(200, 400))

    def fuzz_unblock_user(self):
        users = self.get_users().json()["users"]

        for user in users:
            if user["blocked"]:
                self.block_user(user, comment="Ha!", expected_status=(200, 400))

    def fuzz_update_known_documents(self):
        self.known_documents = self.get_documents().json()["documents"]

    def fuzz_create_document(self):
        self.create_document(data=str(random.randbytes(8)), expected_status=[200, 403])

    def fuzz_modify_document(self):
        if not self.known_documents:
            self.fuzz_update_known_documents()

        doc = random.choice(self.known_documents)

        self.modify_document(
            doc,
            data=str(random.randbytes(8)),
            comment="Hop!",
            expected_status=[200, 400, 403, 409],
        )

    def fuzz_get_document(self):
        if not self.known_documents:
            self.fuzz_update_known_documents()

        doc = random.choice(self.known_documents)

        self.get_document(doc)

    def fuzz_get_version(self):
        if not self.known_documents:
            self.fuzz_update_known_documents()

        doc = random.choice(self.known_documents)

        versions = self.get_versions(doc).json()["versions"]

        version = random.choice(versions)

        self.get_version(version)

    def fuzz_login(self):
        i = random.randint(0, self.session_count - 1)
        self.login_user(f"user_{i}")

    def fuzz_merge_documents(self):
        document_to_merge = random.choice(self.known_documents)
        document_destination = random.choice(self.known_documents)

        self.merge_documents(
            document_to_merge, document_destination, comment="merge it!", expected_status=[200, 400, 403]
        )

    def fuzz_add_tag(self):
        doc = random.choice(self.known_documents)

        chars = string.ascii_letters + "-_" + string.digits
        name = "".join(random.choices(chars, k=random.randint(1, 16)))

        self.add_user_tag(name, doc)

    def fuzz_remove_tag(self):
        tags = self.get_user_tags(user=self.logged_user).json()["user_tags"]

        if len(tags) != 0:
            tag = random.choice(tags)
            self.remove_user_tag(tag["name"], tag["document_id"])

    def possible_actions(self):
        result = [
            (self.fuzz_update_known_documents, 30),
            (self.fuzz_get_document, 50),
            (self.fuzz_get_version, 5),
            (self.get_logs, 1),
        ]

        if self.is_anonymous:
            result += [(self.fuzz_login, 1)]
        else:
            result += [
                (self.get_current_user, 1),
                (self.logout_user, 1),
                (self.fuzz_create_document, 5),
                (self.fuzz_modify_document, 20),
                (self.fuzz_add_tag, 5),
                (self.fuzz_remove_tag, 5),
            ]

        if self.is_moderator:
            result += [
                (self.fuzz_block_user, 1),
                (self.fuzz_unblock_user, 1),
                (self.fuzz_get_user, 1),
                (self.fuzz_merge_documents, 3),
            ]

        return [item[0] for item in result], [item[1] for item in result]

    def do_some_stuff(self):
        possible_actions, weights = self.possible_actions()
        action = random.choices(possible_actions, weights=weights, k=1)[0]

        action()

    def run(self):
        while not self.kill:
            try:
                self.do_some_stuff()
            except:  # pylint: disable=bare-except
                pass


def print_stats(sessions):
    stats = collections.defaultdict(list)
    for session in sessions:
        for key, values in session.stats.items():
            stats[key] += values

    sorted_keys = {}
    for (method, endpoint, outcome), values in stats.items():
        sort_key = f"{endpoint} {method} {outcome}"
        sorted_keys[sort_key] = (method, endpoint, outcome)

        values.sort()

    for sort_key in sorted(sorted_keys):
        (method, endpoint, outcome) = sorted_keys[sort_key]
        values = stats[(method, endpoint, outcome)]

        size = len(values)
        p_90 = int(values[int(size * 0.9)])
        p_95 = int(values[int(size * 0.95)])
        p_99 = int(values[int(size * 0.99)])
        print(f"{method:6} {endpoint:15} {len(values):5} {p_90:5}ms {p_95:5}ms {p_99:5}ms {outcome}")

    print()


def main():

    session_count = 10

    docker_client = docker.from_env()
    containers_stats = list(container.stats(decode=True) for container in docker_client.containers.list(all=True))
    docker_stats = []

    sessions = [FuzzerSession(session_count) for _ in range(session_count)]

    for i, session in enumerate(sessions):
        session.setup_user(f"user_{i}")
        session.create_document(data="init", comment="init")

    sessions[0].login_user("admin")
    sessions[0].add_user_role(2, role="moderator", comment="I trust him")
    sessions[0].add_user_role(3, role="moderator", comment="I trust him")

    threads = [threading.Thread(target=session.run) for session in sessions]

    for thread in threads:
        thread.start()

    for _ in range(50):
        next_wake_up = datetime.now() + timedelta(seconds=1)
        print_stats(sessions)
        docker_stats.append({"docker_stats": list(next(stats) for stats in containers_stats)})
        sleep_time = max(0, (next_wake_up - datetime.now()).total_seconds())
        time.sleep(sleep_time)

    for session in sessions:
        session.kill = True

    for thread in threads:
        thread.join()

    print("#" * 80)
    print_stats(sessions)
    with open("logs/docker_stats.json", mode="w", encoding="utf-8") as f:
        json.dump(docker_stats, f, indent=2)


if __name__ == "__main__":
    main()



import json

from bokeh import io as bokeh_io
from bokeh.io import curdoc
from bokeh.layouts import gridplot
from bokeh.models import (
    ColumnDataSource,
    CrosshairTool,
    LinearAxis,
    NumeralTickFormatter,
    Range1d,
)
from bokeh.models.callbacks import CustomJS
from bokeh.plotting import figure, output_file


def _build_figure(plot_height=300, xaxis=False, **kwargs):
    result = figure(plot_width=1300, plot_height=plot_height, tools="xpan,xwheel_zoom", **kwargs)
    result.xaxis.visible = xaxis

    return result


def _compute_source(filename):

    with open(filename, "r", encoding="utf-8") as f:
        raw_data = [{stat["name"]: stat for stat in item["docker_stats"]} for item in json.load(f)]

    container_names = raw_data[0].keys()

    data = dict(
        x=range(len(raw_data)),
    )

    for name in container_names:
        data[f"{name}_mem"] = [item[name]["memory_stats"]["usage"] for item in raw_data]

        # https://stackoverflow.com/questions/30271942/get-docker-container-cpu-usage-as-percentage
        previous_cpu, previous_cpu_system = 0, 0
        cpu_percent = []
        for item in raw_data:
            cpu_stats = item[name]["cpu_stats"]

            cpu_delta = cpu_stats["cpu_usage"]["total_usage"] - previous_cpu
            system_delta = cpu_stats["system_cpu_usage"] - previous_cpu_system

            if system_delta > 0.0 and cpu_delta > 0.0:
                cpu_percent.append((cpu_delta / system_delta) * cpu_stats["online_cpus"])
            else:
                cpu_percent.append(0)

            previous_cpu, previous_cpu_system = cpu_stats["cpu_usage"]["total_usage"], cpu_stats["system_cpu_usage"]

        data[f"{name}_cpu"] = cpu_percent

    return ColumnDataSource(data=data)


def _build_container_graph(source, containers, x_range=None):
    graph = _build_figure(x_range=x_range)
    # mem_graph.add_tools(HoverTool(tooltips=[("#", "$index"), ("Memory", "@memory")]))

    colors = ("blue", "green", "red")

    graph.extra_y_ranges = {"cpu": Range1d(start=0, end=1)}

    mem_values = []
    for i, container in enumerate(containers):

        mem_source = f"/flask-camp-{container}_mem"
        mem_values += source.data[mem_source]

        graph.line(
            "x",
            mem_source,
            legend_label=f"Memory {container}",
            line_width=2,
            line_color=colors[i],
            source=source,
        )

        graph.line(
            "x",
            f"/flask-camp-{container}_cpu",
            legend_label=f"CPU {container}",
            line_width=2,
            line_dash="dashed",
            line_color=colors[i],
            source=source,
            y_range_name="cpu",
        )

    graph.add_layout(LinearAxis(y_range_name="cpu"), "right")
    graph.y_range = Range1d(start=min(mem_values) * 0.9, end=max(mem_values) * 1.1)
    graph.yaxis[0].formatter = NumeralTickFormatter(format="0 b")
    graph.yaxis[1].formatter = NumeralTickFormatter(format="0%")

    return graph


def _add_crosshair(figures):
    js_move = """
    var start = fig.x_range.start;
    var end = fig.x_range.end;

    if(cb_obj.x>=start && cb_obj.x<=end && cb_obj.y>=start && cb_obj.y<=end) {
        cross.spans.height.computed_location=cb_obj.sx;
    } else {
        cross.spans.height.computed_location = null;
    }
    """

    crosshair = CrosshairTool(dimensions="height", line_color="lightgrey")

    for main_figure in list(figures):
        main_figure.add_tools(crosshair)
        for other_figure in figures:
            if other_figure != main_figure:
                callback = CustomJS(args={"cross": crosshair, "fig": other_figure}, code=js_move)
                main_figure.js_on_event("mouseenter", callback)


def main(filename="logs/docker_stats.json"):
    source = _compute_source(filename)

    haproxy_graph = _build_container_graph(source, ("haproxy-1",))

    figures = (
        haproxy_graph,
        _build_container_graph(source, ("app-1", "app-2", "app-3"), x_range=haproxy_graph.x_range),
        _build_container_graph(source, ("redis-1",), x_range=haproxy_graph.x_range),
        _build_container_graph(source, ("pg-1",), x_range=haproxy_graph.x_range),
    )

    _add_crosshair(figures)

    figures = [(figure,) for figure in figures]
    grid = gridplot(figures, toolbar_location="right")
    curdoc().theme = "dark_minimal"
    output_file(filename + ".html")
    bokeh_io.save(grid)
    bokeh_io.show(grid)


if __name__ == "__main__":
    main()



from flask import request
from werkzeug.exceptions import BadRequest, NotFound

from flask_camp._schemas import schema
from flask_camp._utils import current_api
from flask_camp.models._user import User
from flask_camp._services._security import allow

rule = "/roles/<int:user_id>"


@allow("anonymous", "authenticated")
def get(user_id):
    """Get roles for a given user"""

    user = User.get(id=user_id)

    if user is None:
        raise NotFound()

    return {"status": "ok", "roles": user.roles}


@allow("admin")
@schema("modify_role.json")
def post(user_id):
    """Add a role to an user"""
    user = User.get(id=user_id, with_for_update=True)

    if user is None:
        raise NotFound()

    data = request.get_json()

    role = data["role"]

    if role not in current_api.user_roles:
        raise BadRequest(f"'{role}' doesn't exists. Possible roles are {sorted(current_api.user_roles)}.")

    if role in user.roles:
        raise BadRequest("User has this role")

    user.roles = user.roles + [role]

    current_api.add_log(action=f"add_role {role}", comment=data["comment"], target_user=user)
    current_api.database.session.commit()

    return {"status": "ok", "roles": user.roles}


@allow("admin")
@schema("modify_role.json")
def delete(user_id):
    """Remove a role from an user"""
    user = User.get(id=user_id, with_for_update=True)

    if user is None:
        raise NotFound()

    data = request.get_json()

    role = data["role"]

    if role not in user.roles:
        raise BadRequest("User does not have this role")

    roles = list(user.roles)
    roles.remove(role)
    user.roles = roles

    current_api.add_log(action=f"remove_role {role}", comment=data["comment"], target_user=user)
    current_api.database.session.commit()

    return {"status": "ok", "roles": user.roles}



import pytest
from werkzeug.exceptions import Forbidden

from flask_camp import RestApi
from flask_camp.exceptions import ConfigurationError
from tests.unit_tests.utils import BaseTest


def before_document_delete(user, document_as_dict):
    """Stupid test: user cant delete if it's the last editor"""
    if document_as_dict["user"]["id"] == user.id:
        raise Forbidden()


class Test_errors:
    def test_error(self):
        with pytest.raises(ConfigurationError):
            RestApi(before_document_delete={})


class Test_CustomDelete(BaseTest):
    rest_api_kwargs = {"user_can_delete": True, "before_document_delete": before_document_delete}

    def test_main(self, user, user_2):
        self.login_user(user)

        document = self.create_document().json["document"]

        self.delete_document(document, expected_status=403)

        self.login_user(user_2)
        self.delete_document(document, expected_status=200)


class Test_NormalConf(BaseTest):
    def test_main(self, user, user_2):
        self.login_user(user)
        document = self.create_document().json["document"]
        self.delete_document(document, expected_status=403)

        self.login_user(user_2)
        self.delete_document(document, expected_status=403)



import pytest

from tests.unit_tests.utils import BaseTest
from flask_camp import RestApi, allow
from flask_camp.exceptions import ConfigurationError


class BotModule:
    rule = "/bot"

    @staticmethod
    @allow("bot")
    def get():
        """Here is a custom post, only for bots"""
        return {"hello": "world"}


class Test_Roles(BaseTest):
    rest_api_kwargs = {"user_roles": "bot,contributor"}

    def test_attribution(self, admin, user):
        self.api.add_modules(self.app, BotModule())

        self.login_user(user)
        self.get("/bot", expected_status=403)

        self.login_user(admin)
        self.add_user_role(user, "bot", "it's a good bot")

        self.login_user(user)
        self.get("/bot", expected_status=200)

    def test_errors(self, admin, user):
        self.login_user(admin)
        r = self.add_user_role(user, "imaginary_role", "comment", expected_status=400).json

        message = "'imaginary_role' doesn't exists. Possible roles are ['admin', 'bot', 'contributor', 'moderator']."
        assert r["description"] == message


class Test_Configuration(BaseTest):
    def test_configuration(self):

        api = RestApi(user_roles="bot")
        assert "bot" in api.user_roles

        api = RestApi(user_roles="BOT")
        assert "bot" in api.user_roles

        api = RestApi(user_roles="bot, contributor,")
        assert "bot" in api.user_roles
        assert "contributor" in api.user_roles
        assert "" not in api.user_roles

        api = RestApi(user_roles="")
        assert "" not in api.user_roles

    def test_configuration_errors(self):

        with pytest.raises(ConfigurationError):
            RestApi(user_roles="anonymous")

        with pytest.raises(ConfigurationError):
            RestApi(user_roles="authenticated")



from tests.unit_tests.utils import BaseTest


class Test_RenameUser(BaseTest):
    def test_not_allowed(self, user, user_2):
        self.rename_user(user, "toto", expected_status=403)
        self.rename_user(user_2, "toto", expected_status=403)

        self.login_user(user)

        self.block_user(user, expected_status=403)
        self.unblock_user(user, expected_status=403)

    def test_main(self, user, moderator):
        self.login_user(moderator)

        r = self.get_user(user)
        assert r.json["user"]["name"] == user.name

        self.rename_user(user, "new_name")

        r = self.get_user(user)
        assert r.json["user"]["name"] == "new_name"

    def test_errors(self, user, moderator):
        self.login_user(moderator)

        self.rename_user(42, "toto", expected_status=404)
        self.rename_user(user, "@not_a_good_name", expected_status=400)



import pytest

from flask_camp import RestApi
from tests.unit_tests.utils import BaseTest


class Test_Errors:
    def test_error(self):
        with pytest.raises(FileNotFoundError):
            RestApi(schemas_directory="tests/unit_tests/schemas/", document_schemas=["notfound.json"])

        with pytest.raises(FileNotFoundError):
            RestApi(schemas_directory="tests/not_the_dir/")


class Test_Schemas(BaseTest):
    rest_api_kwargs = {
        "schemas_directory": "tests/unit_tests/schemas",
        "document_schemas": ["outing.json"],
    }

    def test_basic(self, user):

        invalid_data = (
            {"namespace": "outing"},
            {"namespace": "outing", "value": None},
            {"namespace": "outing", "value": 12},
            {"namespace": "outing", "value": "str"},
            {"namespace": "outing", "value": "str", "rating": None},
            {"namespace": "outing", "value": "str", "rating": 12},
            {"namespace": "outing", "value": "str", "rating": "a6"},
        )

        self.login_user(user)

        for data in invalid_data:
            self.create_document(data=data, expected_status=400)

        doc = self.create_document(
            data={"namespace": "outing", "value": "str", "rating": "6a"},
            expected_status=200,
        ).json["document"]

        for data in invalid_data:
            self.modify_document(doc, data=data, expected_status=400)

        self.modify_document(doc, data={"namespace": "outing", "value": "str", "rating": "6b"}, expected_status=200)

        route = self.create_document(data={"namespace": "route"}, expected_status=200).json["document"]
        self.modify_document(route, data=12, expected_status=200)



from flask import request
from werkzeug.exceptions import BadRequest

from flask_camp.models._log import Log
from flask_camp._services._security import allow

rule = "/logs"


@allow("anonymous", "authenticated")
def get():
    """return all logs"""
    limit = request.args.get("limit", default=30, type=int)
    offset = request.args.get("offset", default=0, type=int)

    if not 0 <= limit <= 100:
        raise BadRequest("Limit can't be lower than 0 or higher than 100")

    logs = Log.query.order_by(Log.id.desc()).limit(limit).offset(offset)
    count = Log.query.count()

    logs = [log.as_dict() for log in logs]

    return {"status": "ok", "logs": logs, "count": count}



import pytest

from flask_camp.exceptions import ConfigurationError

from tests.unit_tests.utils import BaseTest


class Test_ClassInstance(BaseTest):
    def test_main(self):
        class CustomModule:
            rule = "/endpoint"

            @self.api.allow("anonymous")
            def get(self):
                return "ok"

        self.api.add_modules(self.app, CustomModule())

        rules = {url_rule.rule for url_rule in self.app.url_map.iter_rules()}

        assert CustomModule.rule in rules, rules

        self.get("/endpoint", expected_status=200)


class Test_Class(BaseTest):
    def test_main(self):
        class CustomModule:
            rule = "/endpoint"

            @staticmethod
            @self.api.allow("anonymous")
            def get():
                return "ok"

        self.api.add_modules(self.app, CustomModule)

        rules = {url_rule.rule for url_rule in self.app.url_map.iter_rules()}

        assert CustomModule.rule in rules, rules

        self.get("/endpoint", expected_status=200)


class Test_Errors(BaseTest):
    def test_missing_allowed(self):
        class CustomModule:
            rule = "/endpoint"

            def get(self):
                pass

        with pytest.raises(ConfigurationError):
            self.api.add_modules(self.app, CustomModule)

    def test_missing_rule(self):
        class CustomModule:
            @self.api.allow("anonymous")
            def get(self):
                pass

        with pytest.raises(ConfigurationError):
            self.api.add_modules(self.app, CustomModule)

    def test_roles_doesnt_exists(self):
        class CustomModule:
            rule = "/endpoint"

            @self.api.allow("not-a-role")
            def get(self):
                pass

        with pytest.raises(ConfigurationError):
            self.api.add_modules(self.app, CustomModule)

    def test_twice(self):
        class CustomModule:
            rule = "/endpoint"

            @self.api.allow("anonymous")
            def get(self):
                pass

        with pytest.raises(AssertionError):
            self.api.add_modules(self.app, CustomModule, CustomModule)



from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import MetaData

# https://flask-sqlalchemy.palletsprojects.com/en/2.x/config/#using-custom-metadata-and-naming-conventions
convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}

metadata = MetaData(naming_convention=convention)
database = SQLAlchemy(metadata=metadata)



from tests.unit_tests.utils import BaseTest


def _assert_log(log, action, user, document_id=None, version_id=None, target_user_id=None):

    assert log["action"] == action
    assert log["user"]["id"] == user.id
    assert log["document_id"] == document_id
    assert log["version_id"] == version_id
    assert log["target_user_id"] == target_user_id


class Test_Logs(BaseTest):
    rest_api_kwargs = {"user_roles": "bot,contributor"}

    def test_anonymous_get(self):
        self.get_logs()

    def test_hide_version(self, moderator):
        self.login_user(moderator)

        doc = self.create_document().json["document"]
        self.modify_document(doc, data="v2")

        self.hide_version(doc)
        self.unhide_version(doc)

        r = self.get_logs()
        assert r.json["count"] == 2, r.json

        logs = r.json["logs"]

        assert logs[-1]["action"] == "hide_version"
        assert logs[-1]["version_id"] == doc["version_id"]
        assert logs[-1]["document_id"] == doc["id"]
        assert logs[-1]["user"]["id"] == moderator.id
        assert logs[-2]["action"] == "unhide_version"
        assert logs[-2]["version_id"] == doc["version_id"]
        assert logs[-2]["document_id"] == doc["id"]
        assert logs[-2]["user"]["id"] == moderator.id

    def test_errors(self):
        self.get_logs(limit=101, expected_status=400)

    def test_typical_scenario(self, user, moderator, admin):

        self.login_user(moderator)
        self.block_user(user)
        self.unblock_user(user)

        doc = self.create_document().json["document"]
        doc_v2 = self.modify_document(doc, data="v2").json["document"]

        self.protect_document(doc)
        self.unprotect_document(doc)

        self.logout_user()
        self.login_user(admin)

        self.add_user_role(user, "moderator", "comment")
        self.remove_user_role(user, "moderator", "comment")
        self.add_user_role(user, "admin", "comment")
        self.add_user_role(user, "bot", "comment")
        self.remove_user_role(user, "admin", "comment")
        self.remove_user_role(user, "bot", "comment")

        self.delete_version(doc_v2)
        self.delete_document(doc)

        r = self.get_logs()
        assert r.json["count"] == 12, r.json

        logs = r.json["logs"]

        _assert_log(logs[-1], "block", moderator, target_user_id=user.id)
        _assert_log(logs[-2], "unblock", moderator, target_user_id=user.id)

        _assert_log(logs[-3], "protect", moderator, document_id=doc["id"])
        _assert_log(logs[-4], "unprotect", moderator, document_id=doc["id"])

        _assert_log(logs[-5], "add_role moderator", admin, target_user_id=user.id)
        _assert_log(logs[-6], "remove_role moderator", admin, target_user_id=user.id)

        _assert_log(logs[-7], "add_role admin", admin, target_user_id=user.id)
        _assert_log(logs[-8], "add_role bot", admin, target_user_id=user.id)

        _assert_log(logs[-9], "remove_role admin", admin, target_user_id=user.id)
        _assert_log(logs[-10], "remove_role bot", admin, target_user_id=user.id)

        _assert_log(logs[-11], "delete_version", admin, document_id=doc["id"], version_id=doc_v2["version_id"])
        _assert_log(logs[-12], "delete_document", admin, document_id=doc["id"])



from datetime import datetime

from flask_login import current_user
from sqlalchemy import Column, ForeignKey, Integer, String, DateTime
from sqlalchemy.orm import relationship

from flask_camp.models._base_model import BaseModel
from flask_camp.models._user import User


class Log(BaseModel):
    __tablename__ = "log"

    timestamp = Column(DateTime, nullable=False)

    user_id = Column(Integer, ForeignKey(User.id), index=True)
    user = relationship(User, foreign_keys=[user_id])

    action = Column(String(32), nullable=False, index=True)
    comment = Column(String(128), nullable=False, index=True)

    target_user_id = Column(Integer, ForeignKey(User.id), index=True)
    target_user = relationship(User, foreign_keys=[target_user_id])

    # no foreign keys: deletion are possible
    document_id = Column(Integer, index=True)
    version_id = Column(Integer, index=True)
    merged_document_id = Column(Integer, index=True)

    def __init__(self, **kwargs):
        kwargs["timestamp"] = datetime.now()
        kwargs["user_id"] = current_user.id

        super().__init__(**kwargs)

    def as_dict(self):
        return {
            "user": self.user.as_dict(),
            "timestamp": self.timestamp.isoformat(),
            "action": self.action,
            "target_user_id": self.target_user_id,
            "target_user": None if self.target_user is None else self.target_user.as_dict(),
            "document_id": self.document_id,
            "merged_document_id": self.merged_document_id,
            "version_id": self.version_id,
        }



from tests.unit_tests.utils import BaseTest


class Test_Documents(BaseTest):
    def test_basic(self, user):
        self.login_user(user)

        r = self.get_documents()
        assert r.json["status"] == "ok"
        assert r.json["count"] == 0
        assert r.json["documents"] == []

        self.create_document(data={"value": "42"})

        r = self.get_documents()
        assert r.json["status"] == "ok"
        assert r.json["count"] == 1
        assert len(r.json["documents"]) == 1
        assert r.json["documents"][0]["data"] == {"value": "42"}

        self.create_document()
        r = self.get_documents()
        assert r.json["status"] == "ok"
        assert r.json["count"] == 2
        assert len(r.json["documents"]) == 2

    def test_offset_limit(self, user):
        self.login_user(user)

        for i in range(110):
            r = self.create_document(data={"value": f"doc {i}"}, expected_status=200)

        r = self.get_documents().json
        assert r["count"] == 110
        assert len(r["documents"]) == 30

        r = self.get_documents(limit=1).json
        assert len(r["documents"]) == 1

        r = self.get_documents(limit=0).json
        assert len(r["documents"]) == 0

        r = self.get_documents(limit=100).json
        assert len(r["documents"]) == 100

        self.get_documents(limit=101, expected_status=400)

        r = self.get_documents(limit="nan").json
        assert len(r["documents"]) == 30

        r = self.get_documents().json
        assert r["documents"][0]["data"]["value"] == "doc 0"
        assert r["documents"][29]["data"]["value"] == "doc 29"

        r = self.get_documents(offset=30).json
        assert r["documents"][0]["data"]["value"] == "doc 30"
        assert r["documents"][29]["data"]["value"] == "doc 59"

        r = self.get_versions().json
        assert r["count"] == 110
        assert len(r["versions"]) == 30

        r = self.get_versions(limit=1).json
        assert len(r["versions"]) == 1

        self.get_versions(limit=101, expected_status=400)



from flask import request
from sqlalchemy import Column, String, ForeignKey, delete

from flask_camp import current_api
from flask_camp.models import BaseModel, Document

from tests.unit_tests.utils import BaseTest


class DocumentSearch(BaseModel):
    id = Column(ForeignKey(Document.id, ondelete="CASCADE"), index=True, nullable=True, primary_key=True)

    document_type = Column(String(16), index=True)


def before_document_save(document):
    if document.last_version is None:  # document as been merged
        delete(DocumentSearch).where(DocumentSearch.id == document.id)
        return

    version = document.last_version

    result = DocumentSearch.get(id=document.id)
    if result is None:
        result = DocumentSearch(id=document.id)
        current_api.database.session.add(result)

    if isinstance(version.data, dict):
        result.document_type = version.data.get("type")


def update_search_query(query):
    document_type = request.args.get("t", default=None, type=str)

    if document_type is not None:
        query = query.join(DocumentSearch).where(DocumentSearch.document_type == document_type)

    return query


class Test_CustomSearch(BaseTest):
    rest_api_kwargs = {
        "before_document_save": before_document_save,
        "update_search_query": update_search_query,
    }

    def test_main(self, admin):
        self.login_user(admin)
        self.add_user_role(admin, "moderator", "I'am god")

        doc_1 = self.create_document(data={"type": "x"}).json["document"]
        doc_2 = self.create_document(data={"type": ""}).json["document"]

        documents = self.get_documents(params={"t": "x"}).json["documents"]
        assert len(documents) == 1
        assert documents[0]["id"] == doc_1["id"]

        self.delete_document(doc_1)
        documents = self.get_documents(params={"t": "x"}).json["documents"]
        assert len(documents) == 0

        doc_2_v2 = self.modify_document(doc_2, data={"type": "x"}).json["document"]
        documents = self.get_documents(params={"t": "x"}).json["documents"]
        assert len(documents) == 1
        assert documents[0]["id"] == doc_2["id"]

        self.hide_version(doc_2_v2)
        documents = self.get_documents(params={"t": "x"}).json["documents"]
        assert len(documents) == 0

        self.unhide_version(doc_2_v2)
        documents = self.get_documents(params={"t": "x"}).json["documents"]
        assert len(documents) == 1
        assert documents[0]["id"] == doc_2["id"]

        self.delete_version(doc_2_v2)
        documents = self.get_documents(params={"t": "x"}).json["documents"]
        assert len(documents) == 0

    def test_merge_1(self, moderator):
        self.login_user(moderator)

        doc_1 = self.create_document(data={"type": "x"}).json["document"]
        doc_2 = self.create_document(data={"type": ""}).json["document"]
        self.merge_documents(doc_1, doc_2, comment="comment")
        documents = self.get_documents(params={"t": "x"}).json["documents"]
        assert len(documents) == 0

    def test_merge_2(self, moderator):
        self.login_user(moderator)

        doc_1 = self.create_document(data={"type": "x"}).json["document"]
        doc_2 = self.create_document(data={"type": ""}).json["document"]
        self.merge_documents(doc_2, doc_1, comment="comment")
        documents = self.get_documents(params={"t": "x"}).json["documents"]
        assert len(documents) == 0

    def test_merge_3(self, moderator):
        self.login_user(moderator)
        doc_1 = self.create_document(data={"type": ""}).json["document"]
        doc_2 = self.create_document(data={"type": "x"}).json["document"]
        self.merge_documents(doc_1, doc_2, comment="comment")
        documents = self.get_documents(params={"t": "x"}).json["documents"]
        assert len(documents) == 1
        assert documents[0]["id"] == doc_2["id"]

    def test_merge_4(self, moderator):
        self.login_user(moderator)
        doc_1 = self.create_document(data={"type": ""}).json["document"]
        doc_2 = self.create_document(data={"type": "x"}).json["document"]
        self.merge_documents(doc_2, doc_1, comment="comment")
        documents = self.get_documents(params={"t": "x"}).json["documents"]
        assert len(documents) == 1
        assert documents[0]["id"] == doc_1["id"]

    def test_merge_5(self, moderator):
        self.login_user(moderator)
        doc_1 = self.create_document(data={"type": "x"}).json["document"]
        doc_2 = self.create_document(data={"type": "x"}).json["document"]
        self.merge_documents(doc_1, doc_2, comment="comment")
        documents = self.get_documents(params={"t": "x"}).json["documents"]
        assert len(documents) == 1
        assert documents[0]["id"] == doc_2["id"]



from flask_login import current_user

from flask_camp._services._security import allow

rule = "/current_user"


@allow("authenticated", allow_blocked=True)
def get():
    """Get the current authenticated user"""
    return {"status": "ok", "user": current_user.as_dict(include_personal_data=True)}



import pytest

from sqlalchemy import Column, ForeignKey, select
from sqlalchemy.orm import relationship
from werkzeug.exceptions import Forbidden

from flask import request
from flask_camp import RestApi, current_api, allow
from flask_camp.models import Document, User, BaseModel
from flask_camp.exceptions import ConfigurationError
from flask_camp.views.content.document import get as get_document_view
from tests.unit_tests.utils import BaseTest


class ProfilePageLink(BaseModel):
    document_id = Column(ForeignKey(Document.id, ondelete="CASCADE"), index=True, nullable=False, unique=True)
    document = relationship(Document, cascade="all,delete")

    user_id = Column(ForeignKey(User.id, ondelete="CASCADE"), index=True, nullable=False, unique=True)
    user = relationship(User, cascade="all,delete")


def before_user_creation(user):
    """
    Two actions:
    1. user must provide a captcha
    2. a document is created, and user id is set to the document's id
    """

    data = request.get_json()
    if "captcha" not in data:
        raise Forbidden()

    user_page = Document.create(
        comment="Creation of user page",
        data={"user_id": user.id},
        author=user,
    )

    current_api.database.session.add(ProfilePageLink(user=user, document=user_page))


class ProfileView:
    rule = "/profile/<string:name>"

    @allow("anonymous")
    def get(self, name):
        query = select(Document.id).join(ProfilePageLink).join(User).where(User.name == name)
        result = current_api.database.session.execute(query)

        return get_document_view(list(result)[0][0])


class Test_Error:
    def test_error(self):
        with pytest.raises(ConfigurationError):
            RestApi(before_user_creation={})


class Test_BeforeUserCreation(BaseTest):
    rest_api_kwargs = {"before_user_creation": before_user_creation}

    def test_main(self):

        self.api.add_modules(self.app, ProfileView())

        self.create_user(expected_status=403)
        user = self.create_user(name="other_user", json={"captcha": 42}, expected_status=200).json["user"]

        documents = self.get_documents(expected_status=200).json["documents"]
        assert "user_id" in documents[0]["data"]
        assert documents[0]["data"]["user_id"] == user["id"]  # check of link
        assert documents[0]["user"]["id"] == user["id"]  # check of page author

        profile = self.get(f"/profile/{user['name']}").json["document"]
        assert profile["data"]["user_id"] == user["id"]



from tests.unit_tests.utils import BaseTest


class Test_Admin(BaseTest):
    def test_right_attribution(self, admin, user):
        self.login_user(user)
        self.add_user_role(user, "moderator", "comment", expected_status=403)

        r = self.get_user(user, expected_status=200)
        assert r.json["user"]["roles"] == []
        self.logout_user()

        self.login_user(admin)
        self.add_user_role(user, "moderator", "comment", expected_status=200)
        r = self.get_user(user, expected_status=200)
        assert r.json["user"]["roles"] == ["moderator"]



from flask import current_app
from flask_mail import Mail, Message


class SendMail(Mail):
    def send_account_creation(self, email, token, user):
        current_app.logger.info("Send registration mail to user %s", user.name)
        message = Message(
            "Welcome to example.com",  # TODO
            recipients=[email],
            body=f"https://example.com?email_token={token}",
            html=f'<a href="https://example.com?email_token={token}">click</a>',
        )

        self.send(message)

    def send_email_change(self, email, token, user):
        current_app.logger.info("Send mail address update mail to user %s", user.name)
        message = Message(
            "Change email",  # TODO
            recipients=[email],
            body=f"https://example.com?email_token={token}",
            html=f'<a href="https://example.com?email_token={token}">click</a>',
        )

        self.send(message)

    def send_login_token(self, email, token, user):
        current_app.logger.info("Send login token mail to user %s", user.name)
        message = Message(
            "Login token email",  # TODO
            recipients=[email],
            body=f"https://example.com?login_token={token}",
            html=f'<a href="https://example.com?login_token={token}">click</a>',
        )

        self.send(message)



from datetime import datetime, timedelta
import json
import secrets

from flask import current_app
from flask_login import current_user
from sqlalchemy import Column, String, Boolean, DateTime
from sqlalchemy.dialects.postgresql import ARRAY
from werkzeug.exceptions import BadRequest, Forbidden, Unauthorized
from werkzeug.security import generate_password_hash, check_password_hash

from flask_camp.models._base_model import BaseModel
from flask_camp._utils import current_api


class AnonymousUser:  # pylint: disable=too-few-public-methods
    id = None

    blocked = False
    is_anonymous = True
    is_authenticated = False
    is_admin = False
    is_moderator = False

    @property
    def roles(self):
        return ["anonymous"]

    # useless ?
    # def get_id(self):
    #     return None


class User(BaseModel):
    __tablename__ = "user_account"  # as user is a reserved word in postgres, we name it user_account

    name = Column(String(64), index=True, unique=True, nullable=False)
    password_hash = Column(String(128), nullable=False)
    _email = Column("email", String(120), index=True, unique=True)

    _email_to_validate = Column("email_to_validate", String(120))
    # linked with email_to_validate, if it's provided, email is validated
    _email_token = Column("email_token", String(64))

    # unique usage token used to login without a password.
    # Useful for user creation and password reset
    _login_token = Column("login_token", String(64))
    _login_token_expiration_date = Column("login_token_expiration_date", DateTime)

    ui_preferences = Column(String, default="{}", nullable=False)

    roles = Column(ARRAY(String(16)), index=True, default=[])

    blocked = Column(Boolean, default=False, nullable=False)

    def __repr__(self):
        return f"<User {self.name}>"

    def _get_private_property(self, property_name):
        # last security fence, it should never happen

        if current_user.id != self.id and not current_user.is_admin and not current_user.is_moderator:
            current_app.logger.error("Unexpected access to user.%s", property_name)
            raise Forbidden()

        return getattr(self, property_name)

    @property
    def email(self):
        return self._get_private_property("_email")

    @property
    def email_is_validated(self):
        # At login, user is anonymous. But we must know if it's validated

        return self._email is not None

    def set_password(self, password):
        current_app.logger.info("Set %s's password", self)
        self.password_hash = generate_password_hash(password)
        self._login_token = None  # disable any login token

    def check_auth(self, password=None, token=None):
        if password is not None:
            return self._check_password(password)

        if token is not None:
            return self._check_login_token(token)

        return False

    def _check_password(self, password):
        if not check_password_hash(self.password_hash, password):
            current_app.logger.info("Check password failed for %s", self)
            return False

        return True

    def _check_login_token(self, login_token):
        if self._login_token is None or login_token is None:
            return False

        if datetime.now() > self._login_token_expiration_date:
            current_app.logger.info("Login token is expired for %s", self)
            return False

        if self._login_token != login_token:
            current_app.logger.error("Login token check fails for %s", self)
            return False

        return True

    @property
    def is_authenticated(self):
        return True

    @property
    def is_active(self):
        return self._email is not None

    def get_id(self):
        return str(self.id)

    def as_dict(self, include_personal_data=False):
        result = {
            "id": self.id,
            "name": self.name,
            "roles": self.roles,
            "blocked": self.blocked,
            "ui_preferences": json.loads(self.ui_preferences) if self.ui_preferences is not None else {},
        }

        if include_personal_data:
            result["email"] = self.email

        return result

    def set_email(self, email):
        if User.get(_email=email) is not None:
            raise BadRequest("A user still exists with this email")

        self._email_to_validate = email
        # each byte is converted to two hex digit, so we need len/2
        self._email_token = secrets.token_hex(int(self.__class__._email_token.type.length / 2))

        current_app.logger.info("Update %s's email", self)

    def send_account_creation_mail(self):
        current_api.mail.send_account_creation(self._email_to_validate, self._email_token, self)

    def send_email_change_mail(self):
        current_api.mail.send_email_change(self._email_to_validate, self._email_token, self)

    def send_login_token_mail(self):
        current_api.mail.send_login_token(self._email, self._login_token, self)

    def validate_email(self, token):

        if self._email_token is None:
            raise BadRequest("There is no email to validate")

        if token != self._email_token:
            raise Unauthorized("Token doesn't match")

        self._email_token = None
        self._email = self._email_to_validate
        self._email_to_validate = None

    def set_login_token(self):
        # each byte is converted to two hex digit, so we need len/2
        self._login_token = secrets.token_hex(int(self.__class__._login_token.type.length / 2))
        self._login_token_expiration_date = datetime.now() + timedelta(hours=1)

    @property
    def is_admin(self):
        return "admin" in self.roles

    @property
    def is_moderator(self):
        return "moderator" in self.roles

    @property
    def login_token_expiration_date(self):
        return self._login_token_expiration_date



from collections import defaultdict

from flask_camp.__main__ import main
from tests.unit_tests.utils import BaseTest


def cli_main(args):
    args = defaultdict(lambda: False, **args)
    main(args)


class Test_CLI(BaseTest):
    def test_main(self):
        with self.app.app_context():
            self.api.database.drop_all()

        cli_main({"init_db": True})
        cli_main({"add_admin": True, "<name>": "admin", "<email>": "admin@email.com", "<password>": "blah"})

        user = self.login_user("admin", password="blah").json["user"]

        assert user["email"] == "admin@email.com"



from functools import wraps
import logging
import json
import os

from flask import request, current_app
from jsonschema import Draft7Validator, RefResolver
from werkzeug.exceptions import BadRequest


class SchemaValidator:
    def __init__(self, base_dir):
        store = {}
        self._validators = {}

        if base_dir[-1] != "/":
            base_dir = f"{base_dir}/"

        if not os.path.isdir(base_dir):
            raise FileNotFoundError(f"{base_dir} is not a directory")

        BASE_URI = "https://schemas/"

        for root, _, files in os.walk(base_dir):
            for file in files:
                if file.endswith(".json"):
                    filename = os.path.join(root, file)

                    with open(filename, encoding="utf8") as file:
                        data = json.load(file)

                    Draft7Validator.check_schema(data)

                    data["$id"] = filename[len(base_dir) :]
                    store[f"{BASE_URI}{filename[len(base_dir):]}"] = data

        for filename, data in store.items():
            resolver = RefResolver(base_uri=BASE_URI, referrer=data, store=store)
            self._validators[filename[len(BASE_URI) :]] = Draft7Validator(
                data, resolver=resolver, format_checker=Draft7Validator.FORMAT_CHECKER
            )

    def validate(self, data, *filenames):
        for filename in filenames:
            validator = self._validators[filename]
            errors = list(validator.iter_errors(data))

            if len(errors) != 0:
                messages = []

                for error in errors:
                    messages.append(f"{error.message} on instance " + "".join([f"[{repr(i)}]" for i in error.path]))

                raise BadRequest("\n".join(messages))

    def schema(self, filename):
        if not self.exists(filename):
            raise FileNotFoundError(f"{filename} does not exists")

        def decorator(real_method):
            @wraps(real_method)
            def wrapper(*args, **kwargs):
                current_app.logger.debug("Validate %s with %s", request.url_rule, filename)
                self.validate(request.get_json(), filename)

                return real_method(*args, **kwargs)

            return wrapper

        return decorator

    def exists(self, filename):
        return filename in self._validators


# expose a decorator for internal schema validation
schema = SchemaValidator(os.path.dirname(__file__)).schema



from contextlib import contextmanager
from time import perf_counter
from sqlalchemy import create_engine


@contextmanager
def catchtime() -> float:
    start = perf_counter()
    yield
    r = perf_counter() - start
    print(f"Execution time: {r:.4f} secs")


db_host = "localhost"
db_port = "5432"
db_user = "www-data"
db_password = "www-data"
db_name = "c2corg"

database_uri = f"postgresql://{db_user}:{db_password}@{db_host}:{db_port}/{db_name}"

engine = create_engine(database_uri)


i = 0
with catchtime():
    for l in engine.execute("SELECT * FROM guidebook.documents_versions ORDER BY id ASC"):
        i += 1

print(i)



from tests.unit_tests.utils import BaseTest


def _assert_conflict_reponse(r):
    result = r.json
    assert "data" in result
    assert "last_version" in result["data"]
    assert "your_version" in result["data"]

    if result["data"]["last_version"] is not None:
        assert result["data"]["your_version"]["version_id"] != result["data"]["last_version"]["version_id"]


class Test_Conflict(BaseTest):
    def test_basic(self, admin):
        self.login_user(admin)

        v0 = self.create_document().json["document"]
        v1 = self.modify_document(v0).json["document"]
        r = self.modify_document(v0, expected_status=409)
        _assert_conflict_reponse(r)

        v2 = self.modify_document(v1).json["document"]
        r = self.modify_document(v0, expected_status=409)
        _assert_conflict_reponse(r)
        r = self.modify_document(v1, expected_status=409)
        _assert_conflict_reponse(r)

        self.delete_version(v1)
        # v1 is availabble in DB. Though, it should raise an error as v2 exists
        r = self.modify_document(v0, expected_status=409)
        _assert_conflict_reponse(r)

        self.delete_version(v2)
        # now, v0 is the last version, so I can modify from it
        self.modify_document(v0, expected_status=200)



from tests.end_tests.utils import ClientSession


if __name__ == "__main__":
    admin = ClientSession(domain="http://localhost:5000")
    admin.login_user("admin")

    moderator = ClientSession(domain="http://localhost:5000")
    moderator.setup_user("moderator")

    admin.add_user_role(moderator.logged_user["id"], role="moderator", comment="I trust him")

    user = ClientSession(domain="http://localhost:5000")
    user.setup_user("user")

    doc = user.create_document().json()["document"]
    admin.delete_document(doc)

    doc = user.create_document().json()["document"]
    user.protect_document(doc, expected_status=403)
    moderator.protect_document(doc)

    user.modify_document(doc, expected_status=403)
    moderator.modify_document(doc, expected_status=200)

    doc_2 = user.create_document().json()["document"]
    user.modify_document(doc_2, expected_status=200)

    anonymous = ClientSession(domain="http://localhost:5000")
    anonymous.get_documents()

    # admin.block_user(user)



import logging
import sys

from flask import Flask
import pytest

from flask_camp import RestApi
from flask_camp.models import User


tested_app = Flask(__name__, static_folder=None)
tested_app.config.update({"TESTING": True, "SECRET_KEY": "not very secret", "SQLALCHEMY_TRACK_MODIFICATIONS": False})
tested_api = RestApi(app=tested_app)


logging.basicConfig(format="%(asctime)s [%(levelname)8s] %(message)s")


def pytest_configure(config):
    if config.getoption("-v") > 1:
        logging.getLogger("sqlalchemy").addHandler(logging.StreamHandler(sys.stdout))
        logging.getLogger("sqlalchemy").setLevel(logging.INFO)

    if not config.option.collectonly:
        # clean previous uncleaned state
        # do not perform this on collect, editors that automatically collect tests on file change
        # may break current test session
        with tested_app.app_context():

            # why not using tested_api.database.drop_all()?
            # because in some case, a table is not known by the ORM
            # for instance, run test A that define a custom table, stop it during execution (the table is not removed)
            # then run only test B. Table defined in test A is not known

            sql = "SELECT table_name FROM information_schema.tables WHERE table_schema = 'public';"
            rows = tested_api.database.session.execute(sql)
            names = [name for name, in rows]
            if len(names) != 0:
                tested_api.database.session.execute(f"DROP TABLE {','.join(names)} CASCADE;")
                tested_api.database.session.commit()

            tested_api.database.create_all()

        tested_api.memory_cache.flushall()


def _db_add_user(name="name", email=None, password="password", validate_email=True, roles=None):

    with tested_app.app_context():
        instance = User(
            name=name,
            roles=roles if isinstance(roles, (list, tuple)) else roles.split(",") if isinstance(roles, str) else [],
        )
        instance.set_password(password)

        instance.set_email(email if email else f"{name}@site.org")

        if validate_email:
            instance.validate_email(instance._email_token)

        tested_api.database.session.add(instance)
        tested_api.database.session.commit()

        result = User(
            id=instance.id,
            name=instance.name,
            _email=instance._email,
            _email_to_validate=instance._email_to_validate,
            _email_token=instance._email_token,
            roles=instance.roles,
        )

    return result


@pytest.fixture()
def admin():
    yield _db_add_user(name="admin", roles="admin")


@pytest.fixture()
def moderator():
    yield _db_add_user(name="moderator", roles="moderator")


@pytest.fixture()
def user():
    yield _db_add_user()


@pytest.fixture()
def unvalidated_user():
    yield _db_add_user(validate_email=False)


@pytest.fixture()
def user_2():
    yield _db_add_user("user_2")



""" Views related to account operations """

from flask import request
from flask_login import login_user, logout_user
from werkzeug.exceptions import Unauthorized

from flask_camp._schemas import schema
from flask_camp._utils import current_api
from flask_camp._services._security import allow
from flask_camp.models._user import User as UserModel


rule = "/login"


@allow("anonymous", "authenticated")
@schema("login_user.json")
def post():
    """Authentificate an user"""
    data = request.get_json()

    name = data["name"]
    password = data.get("password", None)
    token = data.get("token", None)

    user = UserModel.get(name=name)

    if user is None:
        raise Unauthorized(f"User [{name}] does not exists, or password is wrong")

    if not user.email_is_validated:
        raise Unauthorized("User's email is not validated")

    if user.check_auth(password=password, token=token):
        login_user(user)
    else:
        raise Unauthorized(f"User [{name}] does not exists, or password is wrong")

    current_api.database.session.commit()

    return {"status": "ok", "user": user.as_dict(include_personal_data=True)}


@allow("authenticated", allow_blocked=True)
def delete():
    """Logout current user"""
    logout_user()

    return {"status": "ok"}



from tests.unit_tests.utils import BaseTest


def cooker(document, get_document):  # pylint: disable=unused-argument
    data = document.get("data")

    if isinstance(data, dict):
        document["cooked"] = True


class Test_Merge(BaseTest):
    rest_api_kwargs = {"cooker": cooker}

    def test_main(self, moderator, admin):
        self.login_user(moderator)

        v1 = self.create_document(data="v1").json["document"]
        v2 = self.create_document(data="v2").json["document"]

        v3 = self.modify_document(v1, data="v3").json["document"]
        v4 = self.modify_document(v2, data="v4").json["document"]

        doc = self.get_document(v1).json["document"]
        assert doc["id"] == v1["id"]

        self.merge_documents(document_to_merge=v1, document_destination=v2, comment="test")

        versions = self.get_versions(document=v2).json["versions"]
        versions = self.get_versions().json["versions"]

        assert len(versions) == 4

        assert versions[0]["timestamp"] > versions[1]["timestamp"] > versions[2]["timestamp"] > versions[3]["timestamp"]

        assert versions[0]["version_id"] == v4["version_id"]
        assert versions[1]["version_id"] == v3["version_id"]
        assert versions[2]["version_id"] == v2["version_id"]
        assert versions[3]["version_id"] == v1["version_id"]

        r = self.get_document(v1, expected_status=301)
        assert r.headers["Location"] == f"/document/{v2['id']}"

        self.modify_document(v1, data="v3", expected_status=400)
        self.modify_document(v3, data="v3", expected_status=400)

        self.protect_document(v1, expected_status=400)
        self.unprotect_document(v1, expected_status=400)

        self.logout_user()
        self.login_user(admin)

        self.delete_document(v1, expected_status=200)

    def test_forbidden(self, user):
        self.login_user(user)

        v1 = self.create_document(data="v1").json["document"]
        v2 = self.create_document(data="v2").json["document"]

        self.merge_documents(document_to_merge=v1, document_destination=v2, comment="test", expected_status=403)

    def test_common_doc(self, moderator):
        self.login_user(moderator)

        v1 = self.create_document(data="v1").json["document"]

        self.merge_documents(document_to_merge=v1, document_destination=v1, comment="test", expected_status=400)
        self.merge_documents(document_to_merge=v1, document_destination={"id": 42}, comment="test", expected_status=404)

    def test_no_comment(self, moderator):
        self.login_user(moderator)

        doc_1 = self.create_document().json["document"]
        doc_2 = self.create_document().json["document"]

        self.merge_documents(document_to_merge=doc_1, document_destination=doc_2, comment=None, expected_status=400)

    def test_do_not_see_merged_document(self, moderator):
        self.login_user(moderator)

        doc_1 = self.create_document().json["document"]
        doc_2 = self.create_document().json["document"]

        self.merge_documents(document_to_merge=doc_1, document_destination=doc_2, comment="merged")

        r = self.get_documents().json

        assert r["count"] == 1
        assert r["documents"][0]["id"] == doc_2["id"], r["documents"][0]

    def test_dont_modify_redirection(self, moderator):
        self.login_user(moderator)

        doc_1 = self.create_document().json["document"]
        doc_2 = self.create_document().json["document"]

        self.merge_documents(document_to_merge=doc_1, document_destination=doc_2, comment="merged")

        self.modify_document(doc_1, expected_status=400)

    def test_dont_cook_redirections(self, moderator):
        self.login_user(moderator)

        doc_1 = self.create_document(data={"namespace": "cook-me"}).json["document"]
        doc_2 = self.create_document(data={"namespace": "cook-me"}).json["document"]

        self.merge_documents(document_to_merge=doc_1, document_destination=doc_2, comment="merged")

        doc_1 = self.get_document(doc_1, expected_status=301).json["document"]
        doc_2 = self.get_document(doc_2).json["document"]

        assert "cooked" not in doc_1
        assert "cooked" in doc_2

    def test_merge_redirections(self, moderator):
        self.login_user(moderator)

        doc_1 = self.create_document().json["document"]
        doc_2 = self.create_document().json["document"]
        doc_3 = self.create_document().json["document"]

        self.merge_documents(doc_1, doc_2, comment="Doc 1 is now a redirection")

        self.merge_documents(doc_1, doc_3, comment="nope", expected_status=400)
        self.merge_documents(doc_3, doc_1, comment="nope", expected_status=400)
